@@ -36,14 +36,19 @@ struct ssh_bind_struct {
   char *ecdsakey;
   char *dsakey;
   char *rsakey;
+  char *ed25519key;
   ssh_key ecdsa;
   ssh_key dsa;
   ssh_key rsa;
+  ssh_key ed25519;
   char *bindaddr;
   socket_t bindfd;
   unsigned int bindport;
   int blocking;
   int toaccept;
+  bool config_processed;
+  char *config_dir;
+  char *pubkey_accepted_key_types;
 };
 
 struct ssh_poll_handle_struct *ssh_bind_get_poll(struct ssh_bind_struct
