@@ -43,6 +43,7 @@
 #include <openssl/hmac.h>
 #include <openssl/opensslv.h>
 #include <openssl/rand.h>
+
 #include "libcrypto-compat.h"
 
 #ifdef HAVE_OPENSSL_AES_H
@@ -54,21 +55,22 @@
 #include <openssl/des.h>
 #endif
 
-#if (OPENSSL_VERSION_NUMBER<0x00907000L)
-#define OLD_CRYPTO
+#if (defined(HAVE_VALGRIND_VALGRIND_H) && defined(HAVE_OPENSSL_IA32CAP_LOC))
+#include <valgrind/valgrind.h>
+#define CAN_DISABLE_AESNI
 #endif
 
 #include "libssh/crypto.h"
 
-struct ssh_mac_ctx_struct {
-  enum ssh_mac_e mac_type;
-  union {
-    SHACTX sha1_ctx;
-    SHA256CTX sha256_ctx;
-    SHA384CTX sha384_ctx;
-    SHA512CTX sha512_ctx;
-  } ctx;
-};
+#ifdef HAVE_OPENSSL_EVP_KDF_CTX_NEW_ID
+#include <openssl/kdf.h>
+#endif
+
+#ifdef HAVE_OPENSSL_CRYPTO_CTR128_ENCRYPT
+#include <openssl/modes.h>
+#endif
+
+#include "libssh/crypto.h"
 
 static int libcrypto_initialized = 0;
 
@@ -137,7 +139,7 @@ void sha1_final(unsigned char *md, SHACTX c)
     EVP_MD_CTX_destroy(c);
 }
 
-void sha1(unsigned char *digest, int len, unsigned char *hash)
+void sha1(const unsigned char *digest, int len, unsigned char *hash)
 {
     SHACTX c = sha1_init();
     if (c != NULL) {
@@ -229,7 +231,7 @@ void sha256_final(unsigned char *md, SHA256CTX c)
     EVP_MD_CTX_destroy(c);
 }
 
-void sha256(unsigned char *digest, int len, unsigned char *hash)
+void sha256(const unsigned char *digest, int len, unsigned char *hash)
 {
     SHA256CTX c = sha256_init();
     if (c != NULL) {
@@ -267,7 +269,7 @@ void sha384_final(unsigned char *md, SHA384CTX c)
     EVP_MD_CTX_destroy(c);
 }
 
-void sha384(unsigned char *digest, int len, unsigned char *hash)
+void sha384(const unsigned char *digest, int len, unsigned char *hash)
 {
     SHA384CTX c = sha384_init();
     if (c != NULL) {
@@ -305,7 +307,7 @@ void sha512_final(unsigned char *md, SHA512CTX c)
     EVP_MD_CTX_destroy(c);
 }
 
-void sha512(unsigned char *digest, int len, unsigned char *hash)
+void sha512(const unsigned char *digest, int len, unsigned char *hash)
 {
     SHA512CTX c = sha512_init();
     if (c != NULL) {
@@ -343,70 +345,80 @@ void md5_final(unsigned char *md, MD5CTX c)
     EVP_MD_CTX_destroy(c);
 }
 
-ssh_mac_ctx ssh_mac_ctx_init(enum ssh_mac_e type){
-  ssh_mac_ctx ctx = malloc(sizeof(struct ssh_mac_ctx_struct));
-  if (ctx == NULL) {
+#ifdef HAVE_OPENSSL_EVP_KDF_CTX_NEW_ID
+static const EVP_MD *sshkdf_digest_to_md(enum ssh_kdf_digest digest_type)
+{
+    switch (digest_type) {
+    case SSH_KDF_SHA1:
+        return EVP_sha1();
+    case SSH_KDF_SHA256:
+        return EVP_sha256();
+    case SSH_KDF_SHA384:
+        return EVP_sha384();
+    case SSH_KDF_SHA512:
+        return EVP_sha512();
+    }
     return NULL;
-  }
-
-  ctx->mac_type=type;
-  switch(type){
-    case SSH_MAC_SHA1:
-      ctx->ctx.sha1_ctx = sha1_init();
-      return ctx;
-    case SSH_MAC_SHA256:
-      ctx->ctx.sha256_ctx = sha256_init();
-      return ctx;
-    case SSH_MAC_SHA384:
-      ctx->ctx.sha384_ctx = sha384_init();
-      return ctx;
-    case SSH_MAC_SHA512:
-      ctx->ctx.sha512_ctx = sha512_init();
-      return ctx;
-    default:
-      SAFE_FREE(ctx);
-      return NULL;
-  }
 }
 
-void ssh_mac_update(ssh_mac_ctx ctx, const void *data, unsigned long len) {
-  switch(ctx->mac_type){
-    case SSH_MAC_SHA1:
-      sha1_update(ctx->ctx.sha1_ctx, data, len);
-      break;
-    case SSH_MAC_SHA256:
-      sha256_update(ctx->ctx.sha256_ctx, data, len);
-      break;
-    case SSH_MAC_SHA384:
-      sha384_update(ctx->ctx.sha384_ctx, data, len);
-      break;
-    case SSH_MAC_SHA512:
-      sha512_update(ctx->ctx.sha512_ctx, data, len);
-      break;
-    default:
-      break;
-  }
+int ssh_kdf(struct ssh_crypto_struct *crypto,
+            unsigned char *key, size_t key_len,
+            int key_type, unsigned char *output,
+            size_t requested_len)
+{
+    EVP_KDF_CTX *ctx = EVP_KDF_CTX_new_id(EVP_KDF_SSHKDF);
+    int rc;
+
+    if (ctx == NULL) {
+        return -1;
+    }
+
+    rc = EVP_KDF_ctrl(ctx, EVP_KDF_CTRL_SET_MD,
+                      sshkdf_digest_to_md(crypto->digest_type));
+    if (rc != 1) {
+        goto out;
+    }
+    rc = EVP_KDF_ctrl(ctx, EVP_KDF_CTRL_SET_KEY, key, key_len);
+    if (rc != 1) {
+        goto out;
+    }
+    rc = EVP_KDF_ctrl(ctx, EVP_KDF_CTRL_SET_SSHKDF_XCGHASH,
+                      crypto->secret_hash, crypto->digest_len);
+    if (rc != 1) {
+        goto out;
+    }
+    rc = EVP_KDF_ctrl(ctx, EVP_KDF_CTRL_SET_SSHKDF_TYPE, key_type);
+    if (rc != 1) {
+        goto out;
+    }
+    rc = EVP_KDF_ctrl(ctx, EVP_KDF_CTRL_SET_SSHKDF_SESSION_ID,
+                      crypto->session_id, crypto->digest_len);
+    if (rc != 1) {
+        goto out;
+    }
+    rc = EVP_KDF_derive(ctx, output, requested_len);
+    if (rc != 1) {
+        goto out;
+    }
+
+out:
+    EVP_KDF_CTX_free(ctx);
+    if (rc < 0) {
+        return rc;
+    }
+    return 0;
 }
 
-void ssh_mac_final(unsigned char *md, ssh_mac_ctx ctx) {
-  switch(ctx->mac_type){
-    case SSH_MAC_SHA1:
-      sha1_final(md,ctx->ctx.sha1_ctx);
-      break;
-    case SSH_MAC_SHA256:
-      sha256_final(md,ctx->ctx.sha256_ctx);
-      break;
-    case SSH_MAC_SHA384:
-      sha384_final(md,ctx->ctx.sha384_ctx);
-      break;
-    case SSH_MAC_SHA512:
-      sha512_final(md,ctx->ctx.sha512_ctx);
-      break;
-    default:
-      break;
-  }
-  SAFE_FREE(ctx);
+#else
+int ssh_kdf(struct ssh_crypto_struct *crypto,
+            unsigned char *key, size_t key_len,
+            int key_type, unsigned char *output,
+            size_t requested_len)
+{
+    return sshkdf_derive_key(crypto, key, key_len,
+                             key_type, output, requested_len);
 }
+#endif
 
 HMACCTX hmac_init(const void *key, int len, enum ssh_hmac_e type) {
   HMACCTX ctx = NULL;
@@ -416,9 +428,6 @@ HMACCTX hmac_init(const void *key, int len, enum ssh_hmac_e type) {
     return NULL;
   }
 
-#ifndef OLD_CRYPTO
-  HMAC_CTX_reset(ctx); // openssl 0.9.7 requires it.
-#endif
 
   switch(type) {
     case SSH_HMAC_SHA1:
@@ -435,7 +444,6 @@ HMACCTX hmac_init(const void *key, int len, enum ssh_hmac_e type) {
       break;
     default:
       HMAC_CTX_free(ctx);
-      SAFE_FREE(ctx);
       ctx = NULL;
   }
 
@@ -449,14 +457,14 @@ void hmac_update(HMACCTX ctx, const void *data, unsigned long len) {
 void hmac_final(HMACCTX ctx, unsigned char *hashmacbuf, unsigned int *len) {
   HMAC_Final(ctx,hashmacbuf,len);
 
-#ifndef OLD_CRYPTO
+#if OPENSSL_VERSION_NUMBER > 0x10100000L
   HMAC_CTX_free(ctx);
   ctx = NULL;
 #else
   HMAC_cleanup(ctx);
-#endif
-
   SAFE_FREE(ctx);
+  ctx = NULL;
+#endif
 }
 
 static void evp_cipher_init(struct ssh_cipher_struct *cipher) {
@@ -491,13 +499,31 @@ static void evp_cipher_init(struct ssh_cipher_struct *cipher) {
         SSH_LOG(SSH_LOG_WARNING, "This cipher is not available in evp_cipher_init");
         break;
 #endif
+#ifdef HAVE_OPENSSL_EVP_AES_GCM
+    case SSH_AEAD_AES128_GCM:
+        cipher->cipher = EVP_aes_128_gcm();
+        break;
+    case SSH_AEAD_AES256_GCM:
+        cipher->cipher = EVP_aes_256_gcm();
+        break;
+#else
+    case SSH_AEAD_AES128_GCM:
+    case SSH_AEAD_AES256_GCM:
+        SSH_LOG(SSH_LOG_WARNING, "This cipher is not available in evp_cipher_init");
+        break;
+#endif /* HAVE_OPENSSL_EVP_AES_GCM */
     case SSH_3DES_CBC:
         cipher->cipher = EVP_des_ede3_cbc();
         break;
+#ifdef WITH_BLOWFISH_CIPHER
     case SSH_BLOWFISH_CBC:
         cipher->cipher = EVP_bf_cbc();
         break;
         /* ciphers not using EVP */
+#endif
+    case SSH_AEAD_CHACHA20_POLY1305:
+        SSH_LOG(SSH_LOG_WARNING, "The ChaCha cipher cannot be handled here");
+        break;
     case SSH_NO_CIPHER:
         SSH_LOG(SSH_LOG_WARNING, "No valid ciphertype found");
         break;
@@ -517,6 +543,22 @@ static int evp_cipher_set_encrypt_key(struct ssh_cipher_struct *cipher,
         SSH_LOG(SSH_LOG_WARNING, "EVP_EncryptInit_ex failed");
         return SSH_ERROR;
     }
+
+#ifdef HAVE_OPENSSL_EVP_AES_GCM
+    /* For AES-GCM we need to set IV in specific way */
+    if (cipher->ciphertype == SSH_AEAD_AES128_GCM ||
+        cipher->ciphertype == SSH_AEAD_AES256_GCM) {
+        rc = EVP_CIPHER_CTX_ctrl(cipher->ctx,
+                                 EVP_CTRL_GCM_SET_IV_FIXED,
+                                 -1,
+                                 (uint8_t *)IV);
+        if (rc != 1) {
+            SSH_LOG(SSH_LOG_WARNING, "EVP_CTRL_GCM_SET_IV_FIXED failed");
+            return SSH_ERROR;
+        }
+    }
+#endif /* HAVE_OPENSSL_EVP_AES_GCM */
+
     EVP_CIPHER_CTX_set_padding(cipher->ctx, 0);
 
     return SSH_OK;
@@ -534,6 +576,22 @@ static int evp_cipher_set_decrypt_key(struct ssh_cipher_struct *cipher,
         SSH_LOG(SSH_LOG_WARNING, "EVP_DecryptInit_ex failed");
         return SSH_ERROR;
     }
+
+#ifdef HAVE_OPENSSL_EVP_AES_GCM
+    /* For AES-GCM we need to set IV in specific way */
+    if (cipher->ciphertype == SSH_AEAD_AES128_GCM ||
+        cipher->ciphertype == SSH_AEAD_AES256_GCM) {
+        rc = EVP_CIPHER_CTX_ctrl(cipher->ctx,
+                                 EVP_CTRL_GCM_SET_IV_FIXED,
+                                 -1,
+                                 (uint8_t *)IV);
+        if (rc != 1) {
+            SSH_LOG(SSH_LOG_WARNING, "EVP_CTRL_GCM_SET_IV_FIXED failed");
+            return SSH_ERROR;
+        }
+    }
+#endif /* HAVE_OPENSSL_EVP_AES_GCM */
+
     EVP_CIPHER_CTX_set_padding(cipher->ctx, 0);
 
     return SSH_OK;
@@ -541,20 +599,25 @@ static int evp_cipher_set_decrypt_key(struct ssh_cipher_struct *cipher,
 
 /* EVP wrapper function for encrypt/decrypt */
 static void evp_cipher_encrypt(struct ssh_cipher_struct *cipher,
-                        void *in,
-                        void *out,
-                        unsigned long len) {
+                               void *in,
+                               void *out,
+                               size_t len)
+{
     int outlen = 0;
     int rc = 0;
 
-    rc = EVP_EncryptUpdate(cipher->ctx, (unsigned char *)out, &outlen, (unsigned char *)in, len);
+    rc = EVP_EncryptUpdate(cipher->ctx,
+                           (unsigned char *)out,
+                           &outlen,
+                           (unsigned char *)in,
+                           (int)len);
     if (rc != 1){
         SSH_LOG(SSH_LOG_WARNING, "EVP_EncryptUpdate failed");
         return;
     }
     if (outlen != (int)len){
         SSH_LOG(SSH_LOG_WARNING,
-                "EVP_EncryptUpdate: output size %d for %lu in",
+                "EVP_EncryptUpdate: output size %d for %zu in",
                 outlen,
                 len);
         return;
@@ -562,20 +625,25 @@ static void evp_cipher_encrypt(struct ssh_cipher_struct *cipher,
 }
 
 static void evp_cipher_decrypt(struct ssh_cipher_struct *cipher,
-                        void *in,
-                        void *out,
-                        unsigned long len) {
+                               void *in,
+                               void *out,
+                               size_t len)
+{
     int outlen = 0;
     int rc = 0;
 
-    rc = EVP_DecryptUpdate(cipher->ctx, (unsigned char *)out, &outlen, (unsigned char *)in, len);
+    rc = EVP_DecryptUpdate(cipher->ctx,
+                           (unsigned char *)out,
+                           &outlen,
+                           (unsigned char *)in,
+                           (int)len);
     if (rc != 1){
         SSH_LOG(SSH_LOG_WARNING, "EVP_DecryptUpdate failed");
         return;
     }
     if (outlen != (int)len){
         SSH_LOG(SSH_LOG_WARNING,
-                "EVP_DecryptUpdate: output size %d for %lu in",
+                "EVP_DecryptUpdate: output size %d for %zu in",
                 outlen,
                 len);
         return;
@@ -642,10 +710,184 @@ static void aes_ctr_cleanup(struct ssh_cipher_struct *cipher){
 
 #endif /* HAVE_OPENSSL_EVP_AES_CTR */
 
+#ifdef HAVE_OPENSSL_EVP_AES_GCM
+static int
+evp_cipher_aead_get_length(struct ssh_cipher_struct *cipher,
+                           void *in,
+                           uint8_t *out,
+                           size_t len,
+                           uint64_t seq)
+{
+    (void)cipher;
+    (void)seq;
+
+    /* The length is not encrypted: Copy it to the result buffer */
+    memcpy(out, in, len);
+
+    return SSH_OK;
+}
+
+static void
+evp_cipher_aead_encrypt(struct ssh_cipher_struct *cipher,
+                        void *in,
+                        void *out,
+                        size_t len,
+                        uint8_t *tag,
+                        uint64_t seq)
+{
+    size_t authlen, aadlen;
+    uint8_t lastiv[1];
+    int tmplen = 0;
+    size_t outlen;
+    int rc;
+
+    (void) seq;
+
+    aadlen = cipher->lenfield_blocksize;
+    authlen = cipher->tag_size;
+
+    /* increment IV */
+    rc = EVP_CIPHER_CTX_ctrl(cipher->ctx,
+                             EVP_CTRL_GCM_IV_GEN,
+                             1,
+                             lastiv);
+    if (rc == 0) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_CTRL_GCM_IV_GEN failed");
+        return;
+    }
+
+    /* Pass over the authenticated data (not encrypted) */
+    rc = EVP_EncryptUpdate(cipher->ctx,
+                           NULL,
+                           &tmplen,
+                           (unsigned char *)in,
+                           (int)aadlen);
+    outlen = tmplen;
+    if (rc == 0 || outlen != aadlen) {
+        SSH_LOG(SSH_LOG_WARNING, "Failed to pass authenticated data");
+        return;
+    }
+    memcpy(out, in, aadlen);
+
+    /* Encrypt the rest of the data */
+    rc = EVP_EncryptUpdate(cipher->ctx,
+                           (unsigned char *)out + aadlen,
+                           &tmplen,
+                           (unsigned char *)in + aadlen,
+                           (int)len - aadlen);
+    outlen = tmplen;
+    if (rc != 1 || outlen != (int)len - aadlen) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_EncryptUpdate failed");
+        return;
+    }
+
+    /* compute tag */
+    rc = EVP_EncryptFinal(cipher->ctx,
+                          NULL,
+                          &tmplen);
+    if (rc < 0) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_EncryptFinal failed: Failed to create a tag");
+        return;
+    }
+
+    rc = EVP_CIPHER_CTX_ctrl(cipher->ctx,
+                             EVP_CTRL_GCM_GET_TAG,
+                             authlen,
+                             (unsigned char *)tag);
+    if (rc != 1) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_CTRL_GCM_GET_TAG failed");
+        return;
+    }
+}
+
+static int
+evp_cipher_aead_decrypt(struct ssh_cipher_struct *cipher,
+                        void *complete_packet,
+                        uint8_t *out,
+                        size_t encrypted_size,
+                        uint64_t seq)
+{
+    size_t authlen, aadlen;
+    uint8_t lastiv[1];
+    int outlen = 0;
+    int rc = 0;
+
+    (void)seq;
+
+    aadlen = cipher->lenfield_blocksize;
+    authlen = cipher->tag_size;
+
+    /* increment IV */
+    rc = EVP_CIPHER_CTX_ctrl(cipher->ctx,
+                             EVP_CTRL_GCM_IV_GEN,
+                             1,
+                             lastiv);
+    if (rc == 0) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_CTRL_GCM_IV_GEN failed");
+        return SSH_ERROR;
+    }
+
+    /* set tag for authentication */
+    rc = EVP_CIPHER_CTX_ctrl(cipher->ctx,
+                             EVP_CTRL_GCM_SET_TAG,
+                             authlen,
+                             (unsigned char *)complete_packet + aadlen + encrypted_size);
+    if (rc == 0) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_CTRL_GCM_SET_TAG failed");
+        return SSH_ERROR;
+    }
+
+    /* Pass over the authenticated data (not encrypted) */
+    rc = EVP_DecryptUpdate(cipher->ctx,
+                           NULL,
+                           &outlen,
+                           (unsigned char *)complete_packet,
+                           (int)aadlen);
+    if (rc == 0) {
+        SSH_LOG(SSH_LOG_WARNING, "Failed to pass authenticated data");
+        return SSH_ERROR;
+    }
+    /* Do not copy the length to the target buffer, because it is already processed */
+    //memcpy(out, complete_packet, aadlen);
+
+    /* Decrypt the rest of the data */
+    rc = EVP_DecryptUpdate(cipher->ctx,
+                           (unsigned char *)out,
+                           &outlen,
+                           (unsigned char *)complete_packet + aadlen,
+                           encrypted_size /* already substracted aadlen*/);
+    if (rc != 1) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_DecryptUpdate failed");
+        return SSH_ERROR;
+    }
+
+    if (outlen != (int)encrypted_size) {
+        SSH_LOG(SSH_LOG_WARNING,
+                "EVP_DecryptUpdate: output size %d for %zd in",
+                outlen,
+                encrypted_size);
+        return SSH_ERROR;
+    }
+
+    /* verify tag */
+    rc = EVP_DecryptFinal(cipher->ctx,
+                          NULL,
+                          &outlen);
+    if (rc < 0) {
+        SSH_LOG(SSH_LOG_WARNING, "EVP_DecryptFinal failed: Failed authentication");
+        return SSH_ERROR;
+    }
+
+    return SSH_OK;
+}
+
+#endif /* HAVE_OPENSSL_EVP_AES_GCM */
+
 /*
  * The table of supported ciphers
  */
 static struct ssh_cipher_struct ssh_ciphertab[] = {
+#ifdef WITH_BLOWFISH_CIPHER
   {
     .name = "blowfish-cbc",
     .blocksize = 8,
@@ -657,6 +899,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
     .decrypt = evp_cipher_decrypt,
     .cleanup = evp_cipher_cleanup
   },
+#endif
 #ifdef HAS_AES
 #ifndef BROKEN_AES_CTR
 /* OpenSSL until 0.9.7c has a broken AES_ctr128_encrypt implementation which
@@ -665,7 +908,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
 #ifdef HAVE_OPENSSL_EVP_AES_CTR
   {
     .name = "aes128-ctr",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES128_CTR,
     .keysize = 128,
     .set_encrypt_key = evp_cipher_set_encrypt_key,
@@ -676,7 +919,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
   },
   {
     .name = "aes192-ctr",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES192_CTR,
     .keysize = 192,
     .set_encrypt_key = evp_cipher_set_encrypt_key,
@@ -687,7 +930,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
   },
   {
     .name = "aes256-ctr",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES256_CTR,
     .keysize = 256,
     .set_encrypt_key = evp_cipher_set_encrypt_key,
@@ -699,7 +942,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
 #else /* HAVE_OPENSSL_EVP_AES_CTR */
   {
     .name = "aes128-ctr",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES128_CTR,
     .keysize = 128,
     .set_encrypt_key = aes_ctr_set_key,
@@ -710,7 +953,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
   },
   {
     .name = "aes192-ctr",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES192_CTR,
     .keysize = 192,
     .set_encrypt_key = aes_ctr_set_key,
@@ -721,7 +964,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
   },
   {
     .name = "aes256-ctr",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES256_CTR,
     .keysize = 256,
     .set_encrypt_key = aes_ctr_set_key,
@@ -734,7 +977,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
 #endif /* BROKEN_AES_CTR */
   {
     .name = "aes128-cbc",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES128_CBC,
     .keysize = 128,
     .set_encrypt_key = evp_cipher_set_encrypt_key,
@@ -745,7 +988,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
   },
   {
     .name = "aes192-cbc",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES192_CBC,
     .keysize = 192,
     .set_encrypt_key = evp_cipher_set_encrypt_key,
@@ -756,7 +999,7 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
   },
   {
     .name = "aes256-cbc",
-    .blocksize = 16,
+    .blocksize = AES_BLOCK_SIZE,
     .ciphertype = SSH_AES256_CBC,
     .keysize = 256,
     .set_encrypt_key = evp_cipher_set_encrypt_key,
@@ -765,6 +1008,36 @@ static struct ssh_cipher_struct ssh_ciphertab[] = {
     .decrypt = evp_cipher_decrypt,
     .cleanup = evp_cipher_cleanup
   },
+#ifdef HAVE_OPENSSL_EVP_AES_GCM
+  {
+    .name = "aes128-gcm@openssh.com",
+    .blocksize = AES_BLOCK_SIZE,
+    .lenfield_blocksize = 4, /* not encrypted, but authenticated */
+    .ciphertype = SSH_AEAD_AES128_GCM,
+    .keysize = 128,
+    .tag_size = AES_GCM_TAGLEN,
+    .set_encrypt_key = evp_cipher_set_encrypt_key,
+    .set_decrypt_key = evp_cipher_set_decrypt_key,
+    .aead_encrypt = evp_cipher_aead_encrypt,
+    .aead_decrypt_length = evp_cipher_aead_get_length,
+    .aead_decrypt = evp_cipher_aead_decrypt,
+    .cleanup = evp_cipher_cleanup
+  },
+  {
+    .name = "aes256-gcm@openssh.com",
+    .blocksize = AES_BLOCK_SIZE,
+    .lenfield_blocksize = 4, /* not encrypted, but authenticated */
+    .ciphertype = SSH_AEAD_AES256_GCM,
+    .keysize = 256,
+    .tag_size = AES_GCM_TAGLEN,
+    .set_encrypt_key = evp_cipher_set_encrypt_key,
+    .set_decrypt_key = evp_cipher_set_decrypt_key,
+    .aead_encrypt = evp_cipher_aead_encrypt,
+    .aead_decrypt_length = evp_cipher_aead_get_length,
+    .aead_decrypt = evp_cipher_aead_decrypt,
+    .cleanup = evp_cipher_cleanup
+  },
+#endif /* HAVE_OPENSSL_EVP_AES_GCM */
 #endif /* HAS_AES */
 #ifdef HAS_DES
   {
@@ -803,8 +1076,27 @@ int ssh_crypto_init(void)
     if (libcrypto_initialized) {
         return SSH_OK;
     }
-
+    if (SSLeay() != OPENSSL_VERSION_NUMBER){
+        SSH_LOG(SSH_LOG_WARNING, "libssh compiled with %s "
+            "headers, currently running with %s.",
+            OPENSSL_VERSION_TEXT,
+            SSLeay_version(SSLeay())
+        );
+    }
+#ifdef CAN_DISABLE_AESNI
+    /*
+     * disable AES-NI when running within Valgrind, because they generate
+     * too many "uninitialized memory access" false positives
+     */
+    if (RUNNING_ON_VALGRIND){
+        SSH_LOG(SSH_LOG_INFO, "Running within Valgrind, disabling AES-NI");
+        /* Bit #57 denotes AES-NI instruction set extension */
+        OPENSSL_ia32cap &= ~(1LL << 57);
+    }
+#endif
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     OpenSSL_add_all_algorithms();
+#endif
 
     for (i = 0; ssh_ciphertab[i].name != NULL; i++) {
         int cmp;
@@ -833,8 +1125,10 @@ void ssh_crypto_finalize(void)
         return;
     }
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
     EVP_cleanup();
     CRYPTO_cleanup_all_ex_data();
+#endif
 
     libcrypto_initialized = 0;
 }
