@@ -39,6 +39,7 @@
 #ifdef WITH_SERVER
 #include "libssh/server.h"
 #include "libssh/bind.h"
+#include "libssh/bind_config.h"
 #endif
 
 /**
@@ -61,8 +62,11 @@
  *
  * @see ssh_session_connect()
  */
-int ssh_options_copy(ssh_session src, ssh_session *dest) {
+int ssh_options_copy(ssh_session src, ssh_session *dest)
+{
     ssh_session new;
+    struct ssh_iterator *it = NULL;
+    char *id = NULL;
     int i;
 
     if (src == NULL || dest == NULL) {
@@ -74,7 +78,7 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
         return -1;
     }
 
-    if (src->opts.username) {
+    if (src->opts.username != NULL) {
         new->opts.username = strdup(src->opts.username);
         if (new->opts.username == NULL) {
             ssh_free(new);
@@ -82,7 +86,7 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
         }
     }
 
-    if (src->opts.host) {
+    if (src->opts.host != NULL) {
         new->opts.host = strdup(src->opts.host);
         if (new->opts.host == NULL) {
             ssh_free(new);
@@ -90,12 +94,24 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
         }
     }
 
-    if (src->opts.identity) {
-        struct ssh_iterator *it;
+    if (src->opts.bindaddr != NULL) {
+        new->opts.bindaddr = strdup(src->opts.bindaddr);
+        if (new->opts.bindaddr == NULL) {
+            ssh_free(new);
+            return -1;
+        }
+    }
 
+    /* Remove the default identities */
+    for (id = ssh_list_pop_head(char *, new->opts.identity);
+         id != NULL;
+         id = ssh_list_pop_head(char *, new->opts.identity)) {
+        SAFE_FREE(id);
+    }
+    /* Copy the new identities from the source list */
+    if (src->opts.identity != NULL) {
         it = ssh_list_get_iterator(src->opts.identity);
         while (it) {
-            char *id;
             int rc;
 
             id = strdup((char *) it->data);
@@ -114,7 +130,7 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
         }
     }
 
-    if (src->opts.sshdir) {
+    if (src->opts.sshdir != NULL) {
         new->opts.sshdir = strdup(src->opts.sshdir);
         if (new->opts.sshdir == NULL) {
             ssh_free(new);
@@ -122,7 +138,7 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
         }
     }
 
-    if (src->opts.knownhosts) {
+    if (src->opts.knownhosts != NULL) {
         new->opts.knownhosts = strdup(src->opts.knownhosts);
         if (new->opts.knownhosts == NULL) {
             ssh_free(new);
@@ -130,8 +146,16 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
         }
     }
 
+    if (src->opts.global_knownhosts != NULL) {
+        new->opts.global_knownhosts = strdup(src->opts.global_knownhosts);
+        if (new->opts.global_knownhosts == NULL) {
+            ssh_free(new);
+            return -1;
+        }
+    }
+
     for (i = 0; i < 10; i++) {
-        if (src->opts.wanted_methods[i]) {
+        if (src->opts.wanted_methods[i] != NULL) {
             new->opts.wanted_methods[i] = strdup(src->opts.wanted_methods[i]);
             if (new->opts.wanted_methods[i] == NULL) {
                 ssh_free(new);
@@ -140,7 +164,7 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
         }
     }
 
-    if (src->opts.ProxyCommand) {
+    if (src->opts.ProxyCommand != NULL) {
         new->opts.ProxyCommand = strdup(src->opts.ProxyCommand);
         if (new->opts.ProxyCommand == NULL) {
             ssh_free(new);
@@ -155,13 +179,38 @@ int ssh_options_copy(ssh_session src, ssh_session *dest) {
             return -1;
         }
     }
-    new->opts.fd                   = src->opts.fd;
-    new->opts.port                 = src->opts.port;
-    new->opts.timeout              = src->opts.timeout;
-    new->opts.timeout_usec         = src->opts.timeout_usec;
-    new->opts.compressionlevel     = src->opts.compressionlevel;
-    new->common.log_verbosity      = src->common.log_verbosity;
-    new->common.callbacks          = src->common.callbacks;
+
+    if (src->opts.gss_server_identity != NULL) {
+        new->opts.gss_server_identity = strdup(src->opts.gss_server_identity);
+        if (new->opts.gss_server_identity == NULL) {
+            ssh_free(new);
+            return -1;
+        }
+    }
+
+    if (src->opts.gss_client_identity != NULL) {
+        new->opts.gss_client_identity = strdup(src->opts.gss_client_identity);
+        if (new->opts.gss_client_identity == NULL) {
+            ssh_free(new);
+            return -1;
+        }
+    }
+
+    memcpy(new->opts.options_seen, src->opts.options_seen,
+           sizeof(new->opts.options_seen));
+
+    new->opts.fd                    = src->opts.fd;
+    new->opts.port                  = src->opts.port;
+    new->opts.timeout               = src->opts.timeout;
+    new->opts.timeout_usec          = src->opts.timeout_usec;
+    new->opts.compressionlevel      = src->opts.compressionlevel;
+    new->opts.StrictHostKeyChecking = src->opts.StrictHostKeyChecking;
+    new->opts.gss_delegate_creds    = src->opts.gss_delegate_creds;
+    new->opts.flags                 = src->opts.flags;
+    new->opts.nodelay               = src->opts.nodelay;
+    new->opts.config_processed      = src->opts.config_processed;
+    new->common.log_verbosity       = src->common.log_verbosity;
+    new->common.callbacks           = src->common.callbacks;
 
     *dest = new;
 
@@ -174,10 +223,15 @@ int ssh_options_set_algo(ssh_session session,
 {
     char *p = NULL;
 
-    p = ssh_keep_known_algos(algo, list);
+    if (ssh_fips_mode()) {
+        p = ssh_keep_fips_algos(algo, list);
+    } else {
+        p = ssh_keep_known_algos(algo, list);
+    }
+
     if (p == NULL) {
         ssh_set_error(session, SSH_REQUEST_DENIED,
-                "Setting method: no algorithm for method \"%s\" (%s)",
+                "Setting method: no allowed algorithm for method \"%s\" (%s)",
                 ssh_kex_get_description(algo), list);
         return -1;
     }
@@ -200,7 +254,7 @@ int ssh_options_set_algo(ssh_session session,
  *                The hostname or ip address to connect to (const char *).
  *
  *              - SSH_OPTIONS_PORT:
- *                The port to connect to (unsigned int).
+ *                The port to connect to (unsigned int *).
  *
  *              - SSH_OPTIONS_PORT_STR:
  *                The port to connect to (const char *).
@@ -284,54 +338,18 @@ int ssh_options_set_algo(ssh_session session,
  *                The verbosity of the messages. Every log smaller or
  *                equal to verbosity will be shown.
  *                - SSH_LOG_NOLOG: No logging
- *                - SSH_LOG_RARE: Rare conditions or warnings
- *                - SSH_LOG_ENTRY: API-accessible entrypoints
- *                - SSH_LOG_PACKET: Packet id and size
- *                - SSH_LOG_FUNCTIONS: Function entering and leaving
+ *                - SSH_LOG_WARNING: Only warnings
+ *                - SSH_LOG_PROTOCOL: High level protocol information
+ *                - SSH_LOG_PACKET: Lower level protocol infomations, packet level
+ *                - SSH_LOG_FUNCTIONS: Every function path
  *
  *              - SSH_OPTIONS_LOG_VERBOSITY_STR:
- *                Set the session logging verbosity (const char *).\n
- *                \n
- *                The verbosity of the messages. Every log smaller or
- *                equal to verbosity will be shown.
- *                - SSH_LOG_NOLOG: No logging
- *                - SSH_LOG_RARE: Rare conditions or warnings
- *                - SSH_LOG_ENTRY: API-accessible entrypoints
- *                - SSH_LOG_PACKET: Packet id and size
- *                - SSH_LOG_FUNCTIONS: Function entering and leaving
- *                \n
- *                See the corresponding numbers in libssh.h.
- *
- *              - SSH_OPTIONS_AUTH_CALLBACK:
- *                Set a callback to use your own authentication function
- *                (function pointer).
- *
- *              - SSH_OPTIONS_AUTH_USERDATA:
- *                Set the user data passed to the authentication
- *                function (generic pointer).
- *
- *              - SSH_OPTIONS_LOG_CALLBACK:
- *                Set a callback to use your own logging function
- *                (function pointer).
- *
- *              - SSH_OPTIONS_LOG_USERDATA:
- *                Set the user data passed to the logging function
- *                (generic pointer).
- *
- *              - SSH_OPTIONS_STATUS_CALLBACK:
- *                Set a callback to show connection status in realtime
- *                (function pointer).\n
- *                \n
- *                @code
- *                fn(void *arg, float status)
- *                @endcode
- *                \n
- *                During ssh_connect(), libssh will call the callback
- *                with status from 0.0 to 1.0.
- *
- *              - SSH_OPTIONS_STATUS_ARG:
- *                Set the status argument which should be passed to the
- *                status callback (generic pointer).
+ *                Set the session logging verbosity via a
+ *                string that will be converted to a numerical
+ *                value (e.g. "3") and interpreted according
+ *                to the values of
+ *                SSH_OPTIONS_LOG_VERBOSITY above (const
+ *                char *).
  *
  *              - SSH_OPTIONS_CIPHERS_C_S:
  *                Set the symmetric cipher client to server (const char *,
@@ -346,6 +364,14 @@ int ssh_options_set_algo(ssh_session session,
  *                comma-separated list). ex:
  *                "ecdh-sha2-nistp256,diffie-hellman-group14-sha1,diffie-hellman-group1-sha1"
  *
+ *              - SSH_OPTIONS_HMAC_C_S:
+ *                Set the Message Authentication Code algorithm client to server
+ *                (const char *, comma-separated list).
+ *
+ *              - SSH_OPTIONS_HMAC_S_C:
+ *                Set the Message Authentication Code algorithm server to client
+ *                (const char *, comma-separated list).
+ *
  *              - SSH_OPTIONS_HOSTKEYS:
  *                Set the preferred server host key types (const char *,
  *                comma-separated list). ex:
@@ -421,6 +447,24 @@ int ssh_options_set_algo(ssh_session session,
  *                Set it to disable Nagle's Algorithm (TCP_NODELAY) on the
  *                session socket. (int, 0=false)
  *
+ *              - SSH_OPTIONS_PROCESS_CONFIG
+ *                Set it to false to disable automatic processing of per-user
+ *                and system-wide OpenSSH configuration files. LibSSH
+ *                automatically uses these configuration files unless
+ *                you provide it with this option or with different file (bool).
+ *
+ *              - SSH_OPTIONS_REKEY_DATA
+ *                Set the data limit that can be transferred with a single
+ *                key in bytes. RFC 4253 Section 9 recommends 1GB of data, while
+ *                RFC 4344 provides more specific restrictions, that are applied
+ *                automatically. When specified, the lower value will be used.
+ *                (uint64_t, 0=default)
+ *
+ *              - SSH_OPTIONS_REKEY_TIME
+ *                Set the time limit for a session before intializing a rekey
+ *                in seconds. RFC 4253 Section 9 recommends one hour.
+ *                (uint32_t, 0=off)
+ *
  * @param  value The value to set. This is a generic pointer and the
  *               datatype which is used should be set according to the
  *               type set.
@@ -757,7 +801,11 @@ int ssh_options_set(ssh_session session, enum ssh_options_e type,
                 ssh_set_error_invalid(session);
                 return -1;
             } else {
-                p = ssh_keep_known_algos(SSH_HOSTKEYS, v);
+                if (ssh_fips_mode()) {
+                    p = ssh_keep_fips_algos(SSH_HOSTKEYS, v);
+                } else {
+                    p = ssh_keep_known_algos(SSH_HOSTKEYS, v);
+                }
                 if (p == NULL) {
                     ssh_set_error(session, SSH_REQUEST_DENIED,
                         "Setting method: no known public key algorithm (%s)",
@@ -949,6 +997,39 @@ int ssh_options_set(ssh_session session, enum ssh_options_e type,
                 session->opts.nodelay = (*x & 0xff) > 0 ? 1 : 0;
             }
             break;
+        case SSH_OPTIONS_PROCESS_CONFIG:
+            if (value == NULL) {
+                ssh_set_error_invalid(session);
+                return -1;
+            } else {
+                bool *x = (bool *)value;
+                session->opts.config_processed = !(*x);
+            }
+            break;
+        case SSH_OPTIONS_REKEY_DATA:
+            if (value == NULL) {
+                ssh_set_error_invalid(session);
+                return -1;
+            } else {
+                uint64_t *x = (uint64_t *)value;
+                session->opts.rekey_data = *x;
+            }
+            break;
+        case SSH_OPTIONS_REKEY_TIME:
+            if (value == NULL) {
+                ssh_set_error_invalid(session);
+                return -1;
+            } else {
+                uint32_t *x = (uint32_t *)value;
+                if ((*x * 1000) < *x) {
+                    ssh_set_error(session, SSH_REQUEST_DENIED,
+                                  "The provided value (%" PRIu32 ") for rekey"
+                                  " time is too large", *x);
+                    return -1;
+                }
+                session->opts.rekey_time = (*x) * 1000;
+            }
+            break;
         default:
             ssh_set_error(session, SSH_REQUEST_DENIED, "Unknown ssh option %d", type);
             return -1;
@@ -1327,9 +1408,11 @@ int ssh_options_parse_config(ssh_session session, const char *filename) {
       goto out;
   }
   if (filename == NULL) {
-      r = ssh_config_parse_file(session, "/etc/ssh/ssh_config");
+      r = ssh_config_parse_file(session, GLOBAL_CLIENT_CONFIG);
   }
 
+  /* Do not process the default configuration as part of connection again */
+  session->opts.config_processed = true;
 out:
   free(expanded_filename);
   return r;
@@ -1423,6 +1506,30 @@ static int ssh_bind_set_key(ssh_bind sshbind, char **key_loc,
     return 0;
 }
 
+static int ssh_bind_set_algo(ssh_bind sshbind,
+                             enum ssh_kex_types_e algo,
+                             const char *list)
+{
+    char *p = NULL;
+
+    if (ssh_fips_mode()) {
+        p = ssh_keep_fips_algos(algo, list);
+    } else {
+        p = ssh_keep_known_algos(algo, list);
+    }
+    if (p == NULL) {
+        ssh_set_error(sshbind, SSH_REQUEST_DENIED,
+                      "Setting method: no algorithm for method \"%s\" (%s)",
+                      ssh_kex_get_description(algo), list);
+        return -1;
+    }
+
+    SAFE_FREE(sshbind->wanted_methods[algo]);
+    sshbind->wanted_methods[algo] = p;
+
+    return 0;
+}
+
 /**
  * @brief Set options for an SSH server bind.
  *
@@ -1487,6 +1594,53 @@ static int ssh_bind_set_key(ssh_bind sshbind, char **key_loc,
  *                      - SSH_BIND_OPTIONS_IMPORT_KEY:
  *                        Set the Private Key for the server directly (ssh_key)
  *
+ *                      - SSH_BIND_OPTIONS_CIPHERS_C_S:
+ *                        Set the symmetric cipher client to server (const char *,
+ *                        comma-separated list).
+ *
+ *                      - SSH_BIND_OPTIONS_CIPHERS_S_C:
+ *                        Set the symmetric cipher server to client (const char *,
+ *                        comma-separated list).
+ *
+ *                      - SSH_BIND_OPTIONS_KEY_EXCHANGE:
+ *                        Set the key exchange method to be used (const char *,
+ *                        comma-separated list). ex:
+ *                        "ecdh-sha2-nistp256,diffie-hellman-group14-sha1"
+ *
+ *                      - SSH_BIND_OPTIONS_HMAC_C_S:
+ *                        Set the Message Authentication Code algorithm client
+ *                        to server (const char *, comma-separated list).
+ *
+ *                      - SSH_BIND_OPTIONS_HMAC_S_C:
+ *                        Set the Message Authentication Code algorithm server
+ *                        to client (const char *, comma-separated list).
+ *
+ *                      - SSH_BIND_OPTIONS_CONFIG_DIR:
+ *                        Set the directory (const char *, format string)
+ *                        to be used when the "%d" scape is used when providing
+ *                        paths of configuration files to
+ *                        ssh_bind_options_parse_config().
+ *
+ *                      - SSH_BIND_OPTIONS_PROCESS_CONFIG
+ *                        Set it to false to disable automatic processing of
+ *                        system-wide configuration files. LibSSH automatically
+ *                        uses these configuration files otherwise. This
+ *                        option will only have effect if set before any call
+ *                        to ssh_bind_options_parse_config() (bool).
+ *
+ *                      - SSH_BIND_OPTIONS_PUBKEY_ACCEPTED_KEY_TYPES:
+ *                        Set the public key algorithm accepted by the server
+ *                        (const char *, comma-separated list).
+ *
+ *                      - SSH_BIND_OPTIONS_HOSTKEY_ALGORITHMS:
+ *                        Set the list of allowed hostkey signatures algorithms
+ *                        to offer to the client, ordered by preference. This
+ *                        list is used as a filter when creating the list of
+ *                        algorithms to offer to the client: first the list of
+ *                        possible algorithms is created from the list of keys
+ *                        set and then filtered against this list.
+ *                        (const char *, comma-separated list).
+ *
  * @param  value        The value to set. This is a generic pointer and the
  *                      datatype which should be used is described at the
  *                      corresponding value of type above.
@@ -1494,8 +1648,10 @@ static int ssh_bind_set_key(ssh_bind sshbind, char **key_loc,
  * @return              0 on success, < 0 on error, invalid option, or parameter.
  */
 int ssh_bind_options_set(ssh_bind sshbind, enum ssh_bind_options_e type,
-    const void *value) {
+    const void *value)
+{
   char *p, *q;
+  const char *v;
   int i, rc;
 
   if (sshbind == NULL) {
@@ -1530,7 +1686,9 @@ int ssh_bind_options_set(ssh_bind sshbind, enum ssh_bind_options_e type,
                       "without DSA support");
 #endif
               break;
-          case SSH_KEYTYPE_ECDSA:
+          case SSH_KEYTYPE_ECDSA_P256:
+          case SSH_KEYTYPE_ECDSA_P384:
+          case SSH_KEYTYPE_ECDSA_P521:
 #ifdef HAVE_ECC
               bind_key_loc = &sshbind->ecdsa;
               bind_key_path_loc = &sshbind->ecdsakey;
@@ -1592,7 +1750,9 @@ int ssh_bind_options_set(ssh_bind sshbind, enum ssh_bind_options_e type,
                                   "without DSA support");
 #endif
                     break;
-                case SSH_KEYTYPE_ECDSA:
+                case SSH_KEYTYPE_ECDSA_P256:
+                case SSH_KEYTYPE_ECDSA_P384:
+                case SSH_KEYTYPE_ECDSA_P521:
 #ifdef HAVE_ECC
                     bind_key_loc = &sshbind->ecdsa;
 #else
@@ -1717,6 +1877,117 @@ int ssh_bind_options_set(ssh_bind sshbind, enum ssh_bind_options_e type,
         }
       }
       break;
+    case SSH_BIND_OPTIONS_CIPHERS_C_S:
+        v = value;
+        if (v == NULL || v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            if (ssh_bind_set_algo(sshbind, SSH_CRYPT_C_S, v) < 0)
+                return -1;
+        }
+        break;
+    case SSH_BIND_OPTIONS_CIPHERS_S_C:
+        v = value;
+        if (v == NULL || v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            if (ssh_bind_set_algo(sshbind, SSH_CRYPT_S_C, v) < 0)
+                return -1;
+        }
+        break;
+    case SSH_BIND_OPTIONS_KEY_EXCHANGE:
+        v = value;
+        if (v == NULL || v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            rc = ssh_bind_set_algo(sshbind, SSH_KEX, v);
+            if (rc < 0) {
+                return -1;
+            }
+        }
+        break;
+    case SSH_BIND_OPTIONS_HMAC_C_S:
+        v = value;
+        if (v == NULL || v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            if (ssh_bind_set_algo(sshbind, SSH_MAC_C_S, v) < 0)
+                return -1;
+        }
+        break;
+     case SSH_BIND_OPTIONS_HMAC_S_C:
+        v = value;
+        if (v == NULL || v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            if (ssh_bind_set_algo(sshbind, SSH_MAC_S_C, v) < 0)
+                return -1;
+        }
+        break;
+    case SSH_BIND_OPTIONS_CONFIG_DIR:
+        v = value;
+        SAFE_FREE(sshbind->config_dir);
+        if (v == NULL) {
+            break;
+        } else if (v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            sshbind->config_dir = ssh_path_expand_tilde(v);
+            if (sshbind->config_dir == NULL) {
+                ssh_set_error_oom(sshbind);
+                return -1;
+            }
+        }
+        break;
+    case SSH_BIND_OPTIONS_PUBKEY_ACCEPTED_KEY_TYPES:
+        v = value;
+        if (v == NULL || v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            if (ssh_fips_mode()) {
+                p = ssh_keep_fips_algos(SSH_HOSTKEYS, v);
+            } else {
+                p = ssh_keep_known_algos(SSH_HOSTKEYS, v);
+            }
+            if (p == NULL) {
+                ssh_set_error(sshbind, SSH_REQUEST_DENIED,
+                    "Setting method: no known public key algorithm (%s)",
+                     v);
+                return -1;
+            }
+
+            SAFE_FREE(sshbind->pubkey_accepted_key_types);
+            sshbind->pubkey_accepted_key_types = p;
+        }
+        break;
+    case SSH_BIND_OPTIONS_HOSTKEY_ALGORITHMS:
+        v = value;
+        if (v == NULL || v[0] == '\0') {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            rc = ssh_bind_set_algo(sshbind, SSH_HOSTKEYS, v);
+            if (rc < 0) {
+                return -1;
+            }
+        }
+        break;
+    case SSH_BIND_OPTIONS_PROCESS_CONFIG:
+        if (value == NULL) {
+            ssh_set_error_invalid(sshbind);
+            return -1;
+        } else {
+            bool *x = (bool *)value;
+            sshbind->config_processed = !(*x);
+        }
+        break;
     default:
       ssh_set_error(sshbind, SSH_REQUEST_DENIED, "Unknown ssh option %d", type);
       return -1;
@@ -1725,6 +1996,129 @@ int ssh_bind_options_set(ssh_bind sshbind, enum ssh_bind_options_e type,
 
   return 0;
 }
+
+static char *ssh_bind_options_expand_escape(ssh_bind sshbind, const char *s)
+{
+    char buf[MAX_BUF_SIZE];
+    char *r, *x = NULL;
+    const char *p;
+    size_t i, l;
+
+    r = ssh_path_expand_tilde(s);
+    if (r == NULL) {
+        ssh_set_error_oom(sshbind);
+        return NULL;
+    }
+
+    if (strlen(r) > MAX_BUF_SIZE) {
+        ssh_set_error(sshbind, SSH_FATAL, "string to expand too long");
+        free(r);
+        return NULL;
+    }
+
+    p = r;
+    buf[0] = '\0';
+
+    for (i = 0; *p != '\0'; p++) {
+        if (*p != '%') {
+            buf[i] = *p;
+            i++;
+            if (i >= MAX_BUF_SIZE) {
+                free(r);
+                return NULL;
+            }
+            buf[i] = '\0';
+            continue;
+        }
+
+        p++;
+        if (*p == '\0') {
+            break;
+        }
+
+        switch (*p) {
+            case 'd':
+                x = strdup(sshbind->config_dir);
+                break;
+            default:
+                ssh_set_error(sshbind, SSH_FATAL,
+                        "Wrong escape sequence detected");
+                free(r);
+                return NULL;
+        }
+
+        if (x == NULL) {
+            ssh_set_error_oom(sshbind);
+            free(r);
+            return NULL;
+        }
+
+        i += strlen(x);
+        if (i >= MAX_BUF_SIZE) {
+            ssh_set_error(sshbind, SSH_FATAL,
+                    "String too long");
+            free(x);
+            free(r);
+            return NULL;
+        }
+        l = strlen(buf);
+        strncpy(buf + l, x, sizeof(buf) - l - 1);
+        buf[i] = '\0';
+        SAFE_FREE(x);
+    }
+
+    free(r);
+    return strdup(buf);
+}
+
+/**
+ * @brief Parse a ssh bind options configuration file.
+ *
+ * This parses the options file and set them to the ssh_bind handle provided. If
+ * an option was previously set, it is overridden. If the global configuration
+ * hasn't been processed yet, it is processed prior to the provided file.
+ *
+ * @param  sshbind      SSH bind handle
+ *
+ * @param  filename     The options file to use; if NULL only the global
+ *                      configuration is parsed and applied (if it haven't been
+ *                      processed before).
+ *
+ * @return 0 on success, < 0 on error.
+ */
+int ssh_bind_options_parse_config(ssh_bind sshbind, const char *filename)
+{
+    int rc = 0;
+    char *expanded_filename;
+
+    if (sshbind == NULL) {
+        return -1;
+    }
+
+    /* If the global default configuration hasn't been processed yet, process it
+     * before the provided configuration. */
+    if (!(sshbind->config_processed)) {
+        rc = ssh_bind_config_parse_file(sshbind, GLOBAL_BIND_CONFIG);
+        if (rc != 0) {
+            return rc;
+        }
+        sshbind->config_processed = true;
+    }
+
+    if (filename != NULL) {
+        expanded_filename = ssh_bind_options_expand_escape(sshbind, filename);
+        if (expanded_filename == NULL) {
+            return -1;
+        }
+
+        /* Apply the user provided configuration */
+        rc = ssh_bind_config_parse_file(sshbind, expanded_filename);
+        free(expanded_filename);
+    }
+
+    return rc;
+}
+
 #endif
 
 /** @} */
