@@ -47,6 +47,8 @@
 #include "libssh/kex.h"
 #include "libssh/auth.h"
 #include "libssh/gssapi.h"
+#include "libssh/bytearray.h"
+#include "libssh/dh.h"
 
 static ssh_packet_callback default_packet_handlers[]= {
   ssh_packet_disconnect_callback,          // SSH2_MSG_DISCONNECT                 1
@@ -72,7 +74,7 @@ static ssh_packet_callback default_packet_handlers[]= {
 #else
   NULL,
 #endif
-  ssh_packet_dh_reply,                     // SSH2_MSG_KEXDH_REPLY                31
+  NULL,                                    // SSH2_MSG_KEXDH_REPLY                31
                                            // SSH2_MSG_KEX_DH_GEX_GROUP           31
   NULL,                                    // SSH2_MSG_KEX_DH_GEX_INIT            32
   NULL,                                    // SSH2_MSG_KEX_DH_GEX_REPLY           33
@@ -291,6 +293,7 @@ static enum ssh_packet_filter_result_e ssh_packet_incoming_filter(ssh_session se
          * - session_state == SSH_SESSION_STATE_AUTHENTICATED
          *   or session_state == SSH_SESSION_STATE_INITIAL_KEX
          * - dh_handshake_state == DH_STATE_INIT
+         *   or dh_handshake_state == DH_STATE_INIT_SENT (re-exchange)
          *   or dh_handshake_state == DH_STATE_FINISHED (re-exchange)
          *
          * Transitions:
@@ -309,6 +312,7 @@ static enum ssh_packet_filter_result_e ssh_packet_incoming_filter(ssh_session se
         }
 
         if ((session->dh_handshake_state != DH_STATE_INIT) &&
+            (session->dh_handshake_state != DH_STATE_INIT_SENT) &&
             (session->dh_handshake_state != DH_STATE_FINISHED))
         {
             rc = SSH_PACKET_DENIED;
@@ -384,6 +388,7 @@ static enum ssh_packet_filter_result_e ssh_packet_incoming_filter(ssh_session se
          * States required:
          * - session_state == SSH_SESSION_STATE_DH
          * - dh_handshake_state == DH_STATE_INIT_SENT
+         *   or dh_handshake_state == DH_STATE_REQUEST_SENT (dh-gex)
          *
          * Transitions:
          * - session->dh_handhsake_state = DH_STATE_NEWKEYS_SENT
@@ -394,7 +399,8 @@ static enum ssh_packet_filter_result_e ssh_packet_incoming_filter(ssh_session se
             break;
         }
 
-        if (session->dh_handshake_state != DH_STATE_INIT_SENT) {
+        if (session->dh_handshake_state != DH_STATE_INIT_SENT &&
+            session->dh_handshake_state != DH_STATE_REQUEST_SENT) {
             rc = SSH_PACKET_DENIED;
             break;
         }
@@ -924,6 +930,100 @@ end:
     return rc;
 }
 
+/* Returns current_crypto structure from the session.
+ * During key exchange (or rekey), after one of the sides
+ * sending NEWKEYS packet, this might return next_crypto for one
+ * of the directions that is ahead to send already queued packets
+ */
+struct ssh_crypto_struct *
+ssh_packet_get_current_crypto(ssh_session session,
+                              enum ssh_crypto_direction_e direction)
+{
+    if (session == NULL) {
+        return NULL;
+    }
+
+    if (session->current_crypto != NULL &&
+        session->current_crypto->used & direction) {
+        return session->current_crypto;
+    }
+
+    if (session->next_crypto != NULL &&
+        session->next_crypto->used & direction) {
+        return session->next_crypto;
+    }
+
+    return NULL;
+}
+
+#define MAX_PACKETS    (1UL<<31)
+
+static bool ssh_packet_need_rekey(ssh_session session,
+                                  const uint32_t payloadsize)
+{
+    bool data_rekey_needed = false;
+    struct ssh_crypto_struct *crypto = NULL;
+    struct ssh_cipher_struct *out_cipher = NULL, *in_cipher = NULL;
+    uint32_t next_blocks;
+
+    /* We can safely rekey only in authenticated state */
+    if ((session->flags & SSH_SESSION_FLAG_AUTHENTICATED) == 0) {
+        return false;
+    }
+
+    /* Do not rekey if the rekey/key-exchange is in progress */
+    if (session->dh_handshake_state != DH_STATE_FINISHED) {
+        return false;
+    }
+
+    crypto = ssh_packet_get_current_crypto(session, SSH_DIRECTION_BOTH);
+    if (crypto == NULL) {
+        return false;
+    }
+
+    out_cipher = crypto->out_cipher;
+    in_cipher = crypto->in_cipher;
+
+    /* Make sure we can send at least something for very small limits */
+    if ((out_cipher->packets == 0) && (in_cipher->packets == 0)) {
+        return false;
+    }
+
+    /* Time based rekeying */
+    if (session->opts.rekey_time != 0 &&
+        ssh_timeout_elapsed(&session->last_rekey_time,
+                            session->opts.rekey_time)) {
+        return true;
+    }
+
+    /* RFC4344, Section 3.1 Recommends rekeying after 2^31 packets in either
+     * direction to avoid possible information leakage through the MAC tag
+     */
+    if (out_cipher->packets > MAX_PACKETS ||
+        in_cipher->packets > MAX_PACKETS) {
+        return true;
+    }
+
+    /* Data-based rekeying:
+     *  * For outgoing packets we can still delay them
+     *  * Incoming packets need to be processed anyway, but we can
+     *    signalize our intention to rekey
+     */
+    next_blocks = payloadsize / out_cipher->blocksize;
+    data_rekey_needed = (out_cipher->max_blocks != 0 &&
+                         out_cipher->blocks + next_blocks > out_cipher->max_blocks) ||
+                         (in_cipher->max_blocks != 0 &&
+                         in_cipher->blocks + next_blocks > in_cipher->max_blocks);
+
+    SSH_LOG(SSH_LOG_PACKET,
+            "packet: [data_rekey_needed=%d, out_blocks=%" PRIu64 ", in_blocks=%" PRIu64,
+            data_rekey_needed,
+            out_cipher->blocks + next_blocks,
+            in_cipher->blocks + next_blocks);
+
+    return data_rekey_needed;
+}
+
 /* in nonblocking mode, socket_read will read as much as it can, and return */
 /* SSH_OK if it has read at least len bytes, otherwise, SSH_AGAIN. */
 /* in blocking mode, it will read at least len bytes and will block until it's ok. */
@@ -938,11 +1038,9 @@ end:
  */
 int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
 {
-    ssh_session session= (ssh_session) user;
-    unsigned int blocksize = (session->current_crypto ?
-                              session->current_crypto->in_cipher->blocksize : 8);
-    unsigned int lenfield_blocksize = (session->current_crypto ?
-                                  session->current_crypto->in_cipher->lenfield_blocksize : 8);
+    ssh_session session = (ssh_session)user;
+    unsigned int blocksize = 8;
+    unsigned int lenfield_blocksize = 8;
     size_t current_macsize = 0;
     uint8_t *ptr = NULL;
     int to_be_read;
@@ -955,11 +1053,27 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
     uint8_t padding;
     size_t processed = 0; /* number of byte processed from the callback */
     enum ssh_packet_filter_result_e filter_result;
+    struct ssh_crypto_struct *crypto = NULL;
+    bool etm = false;
+    int etm_packet_offset = 0;
+    bool ok;
 
-    if(session->current_crypto != NULL) {
-      current_macsize = hmac_digest_len(session->current_crypto->in_hmac);
+    crypto = ssh_packet_get_current_crypto(session, SSH_DIRECTION_IN);
+    if (crypto != NULL) {
+        current_macsize = hmac_digest_len(crypto->in_hmac);
+        blocksize = crypto->in_cipher->blocksize;
+        lenfield_blocksize = crypto->in_cipher->lenfield_blocksize;
+        etm = crypto->in_hmac_etm;
     }
-    if (lenfield_blocksize == 0) {
+
+    if (etm) {
+        /* In EtM mode packet size is unencrypted. This means
+         * we need to use this offset and set the block size
+         * that is part of the encrypted part to 0.
+         */
+        etm_packet_offset = sizeof(uint32_t);
+        lenfield_blocksize = 0;
+    } else if (lenfield_blocksize == 0) {
         lenfield_blocksize = blocksize;
     }
     if (data == NULL) {
@@ -982,10 +1096,10 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
 #endif
     switch(session->packet_state) {
         case PACKET_STATE_INIT:
-            if (receivedlen < lenfield_blocksize) {
+            if (receivedlen < lenfield_blocksize + etm_packet_offset) {
                 /*
-                 * We didn't receive enough data to read at least one
-                 * block size, give up
+                 * We didn't receive enough data to read either at least one
+                 * block size or the unencrypted length in EtM mode.
                  */
 #ifdef DEBUG_PACKET
                 SSH_LOG(SSH_LOG_PACKET,
@@ -1012,13 +1126,20 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
                 }
             }
 
-            ptr = ssh_buffer_allocate(session->in_buffer, lenfield_blocksize);
-            if (ptr == NULL) {
-                goto error;
+            if (!etm) {
+                ptr = ssh_buffer_allocate(session->in_buffer, lenfield_blocksize);
+                if (ptr == NULL) {
+                    goto error;
+                }
+                packet_len = ssh_packet_decrypt_len(session, ptr, (uint8_t *)data);
+                to_be_read = packet_len - lenfield_blocksize + sizeof(uint32_t);
+            } else {
+                /* Length is unencrypted in case of Encrypt-then-MAC */
+                packet_len = PULL_BE_U32(data, 0);
+                to_be_read = packet_len - etm_packet_offset;
             }
-            processed += lenfield_blocksize;
-            packet_len = ssh_packet_decrypt_len(session, ptr, (uint8_t *)data);
 
+            processed += lenfield_blocksize + etm_packet_offset;
             if (packet_len > MAX_PACKET_LEN) {
                 ssh_set_error(session,
                               SSH_FATAL,
@@ -1026,7 +1147,6 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
                               packet_len, packet_len);
                 goto error;
             }
-            to_be_read = packet_len - lenfield_blocksize + sizeof(uint32_t);
             if (to_be_read < 0) {
                 /* remote sshd sends invalid sizes? */
                 ssh_set_error(session,
@@ -1041,7 +1161,7 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
             FALL_THROUGH;
         case PACKET_STATE_SIZEREAD:
             packet_len = session->in_packet.len;
-            processed = lenfield_blocksize;
+            processed = lenfield_blocksize + etm_packet_offset;
             to_be_read = packet_len + sizeof(uint32_t) + current_macsize;
             /* if to_be_read is zero, the whole packet was blocksize bytes. */
             if (to_be_read != 0) {
@@ -1056,45 +1176,86 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
                     return 0;
                 }
 
-                packet_second_block = (uint8_t*)data + lenfield_blocksize;
+                packet_second_block = (uint8_t*)data + lenfield_blocksize + etm_packet_offset;
                 processed = to_be_read - current_macsize;
             }
 
             /* remaining encrypted bytes from the packet, MAC not included */
             packet_remaining =
-                packet_len - (lenfield_blocksize - sizeof(uint32_t));
+                packet_len - (lenfield_blocksize - sizeof(uint32_t) + etm_packet_offset);
             cleartext_packet = ssh_buffer_allocate(session->in_buffer,
                                                    packet_remaining);
-            if (session->current_crypto) {
-                /*
-                 * Decrypt the rest of the packet (lenfield_blocksize bytes already
-                 * have been decrypted)
-                 */
-                if (packet_remaining > 0) {
-                    rc = ssh_packet_decrypt(session,
-                                            cleartext_packet,
-                                            (uint8_t *)data,
-                                            lenfield_blocksize,
-                                            processed - lenfield_blocksize);
-                    if (rc < 0) {
-                        ssh_set_error(session, SSH_FATAL, "Decryption error");
-                        goto error;
-                    }
-                }
-                mac = packet_second_block + packet_remaining;
-
-                rc = ssh_packet_hmac_verify(session, session->in_buffer, mac, session->current_crypto->in_hmac);
-                if (rc < 0) {
-                    ssh_set_error(session, SSH_FATAL, "HMAC error");
-                    goto error;
-                }
-                processed += current_macsize;
-            } else {
-                memcpy(cleartext_packet, packet_second_block, packet_remaining);
+            if (cleartext_packet == NULL) {
+                goto error;
             }
 
-            /* skip the size field which has been processed before */
-            ssh_buffer_pass_bytes(session->in_buffer, sizeof(uint32_t));
+            if (packet_second_block != NULL) {
+                if (crypto != NULL) {
+                    mac = packet_second_block + packet_remaining;
+
+                    if (etm) {
+                        rc = ssh_packet_hmac_verify(session,
+                                                    data,
+                                                    processed,
+                                                    mac,
+                                                    crypto->in_hmac);
+                        if (rc < 0) {
+                            ssh_set_error(session, SSH_FATAL, "HMAC error");
+                            goto error;
+                        }
+                    }
+                    /*
+                     * Decrypt the packet. In case of EtM mode, the length is already
+                     * known as it's unencrypted. In the other case, lenfield_blocksize bytes
+                     * already have been decrypted.
+                     */
+                    if (packet_remaining > 0) {
+                        rc = ssh_packet_decrypt(session,
+                                                cleartext_packet,
+                                                (uint8_t *)data,
+                                                lenfield_blocksize + etm_packet_offset,
+                                                processed - (lenfield_blocksize + etm_packet_offset));
+                        if (rc < 0) {
+                            ssh_set_error(session,
+                                          SSH_FATAL,
+                                          "Decryption error");
+                            goto error;
+                        }
+                    }
+
+                    if (!etm) {
+                        rc = ssh_packet_hmac_verify(session,
+                                                    ssh_buffer_get(session->in_buffer),
+                                                    ssh_buffer_get_len(session->in_buffer),
+                                                    mac,
+                                                    crypto->in_hmac);
+                        if (rc < 0) {
+                            ssh_set_error(session, SSH_FATAL, "HMAC error");
+                            goto error;
+                        }
+                    }
+                    processed += current_macsize;
+                } else {
+                    memcpy(cleartext_packet,
+                           packet_second_block,
+                           packet_remaining);
+                }
+            }
+
+#ifdef WITH_PCAP
+            if (session->pcap_ctx != NULL) {
+                ssh_pcap_context_write(session->pcap_ctx,
+                                       SSH_PCAP_DIR_IN,
+                                       ssh_buffer_get(session->in_buffer),
+                                       ssh_buffer_get_len(session->in_buffer),
+                                       ssh_buffer_get_len(session->in_buffer));
+            }
+#endif
+
+            if (!etm) {
+                /* skip the size field which has been processed before */
+                ssh_buffer_pass_bytes(session->in_buffer, sizeof(uint32_t));
+            }
 
             rc = ssh_buffer_get_u8(session->in_buffer, &padding);
             if (rc == 0) {
@@ -1116,8 +1277,7 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
             compsize = ssh_buffer_get_len(session->in_buffer);
 
 #ifdef WITH_ZLIB
-            if (session->current_crypto
-                && session->current_crypto->do_compress_in
+            if (crypto && crypto->do_compress_in
                 && ssh_buffer_get_len(session->in_buffer) > 0) {
                 rc = decompress_buffer(session, session->in_buffer,MAX_PACKET_LEN);
                 if (rc < 0) {
@@ -1127,6 +1287,13 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
 #endif /* WITH_ZLIB */
             payloadsize = ssh_buffer_get_len(session->in_buffer);
             session->recv_seq++;
+            if (crypto != NULL) {
+                struct ssh_cipher_struct *cipher = NULL;
+
+                cipher = crypto->in_cipher;
+                cipher->packets++;
+                cipher->blocks += payloadsize / cipher->blocksize;
+            }
             if (session->raw_counter != NULL) {
                 session->raw_counter->in_bytes += payloadsize;
                 session->raw_counter->in_packets++;
@@ -1151,6 +1318,10 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
                 ssh_packet_process(session, session->in_packet.type);
                 break;
             case SSH_PACKET_DENIED:
+                ssh_set_error(session,
+                              SSH_FATAL,
+                              "Packet filter: rejected packet (type %d)",
+                              session->in_packet.type);
                 goto error;
             case SSH_PACKET_UNKNOWN:
                 ssh_packet_send_unimplemented(session, session->recv_seq - 1);
@@ -1170,6 +1341,16 @@ int ssh_packet_socket_callback(const void *data, size_t receivedlen, void *user)
                 processed += rc;
             }
 
+            ok = ssh_packet_need_rekey(session, 0);
+            if (ok) {
+                SSH_LOG(SSH_LOG_PACKET, "Incoming packet triggered rekey");
+                rc = ssh_send_rekex(session);
+                if (rc != SSH_OK) {
+                    SSH_LOG(SSH_LOG_PACKET, "Rekey failed: rc = %d", rc);
+                    return rc;
+                }
+            }
+
             return processed;
         case PACKET_STATE_PROCESSING:
             SSH_LOG(SSH_LOG_PACKET, "Nested packet processing. Delaying.");
@@ -1231,6 +1412,17 @@ void ssh_packet_set_callbacks(ssh_session session, ssh_packet_callbacks callback
   }
 }
 
+/** @internal
+ * @brief remove the callbacks from the packet layer
+ */
+void ssh_packet_remove_callbacks(ssh_session session, ssh_packet_callbacks callbacks){
+    struct ssh_iterator *it = NULL;
+    it = ssh_list_find(session->packet_callbacks, callbacks);
+    if (it != NULL) {
+        ssh_list_remove(session->packet_callbacks, it);
+    }
+}
+
 /** @internal
  * @brief sets the default packet handlers
  */
@@ -1246,37 +1438,54 @@ void ssh_packet_set_default_callbacks(ssh_session session){
  * @brief dispatch the call of packet handlers callbacks for a received packet
  * @param type type of packet
  */
-void ssh_packet_process(ssh_session session, uint8_t type){
-	struct ssh_iterator *i;
-	int r=SSH_PACKET_NOT_USED;
-	ssh_packet_callbacks cb;
+void ssh_packet_process(ssh_session session, uint8_t type)
+{
+    struct ssh_iterator *i = NULL;
+    int rc = SSH_PACKET_NOT_USED;
+    ssh_packet_callbacks cb;
 
-	SSH_LOG(SSH_LOG_PACKET, "Dispatching handler for packet type %d",type);
-	if(session->packet_callbacks == NULL){
-		SSH_LOG(SSH_LOG_RARE,"Packet callback is not initialized !");
+    SSH_LOG(SSH_LOG_PACKET, "Dispatching handler for packet type %d", type);
+    if (session->packet_callbacks == NULL) {
+        SSH_LOG(SSH_LOG_RARE, "Packet callback is not initialized !");
+        return;
+    }
 
-		return;
-	}
-	i=ssh_list_get_iterator(session->packet_callbacks);
-	while(i != NULL){
-		cb=ssh_iterator_value(ssh_packet_callbacks,i);
-		i=i->next;
-		if(!cb)
-			continue;
-		if(cb->start > type)
-			continue;
-		if(cb->start + cb->n_callbacks <= type)
-			continue;
-		if(cb->callbacks[type - cb->start]==NULL)
-			continue;
-		r=cb->callbacks[type - cb->start](session,type,session->in_buffer,cb->user);
-		if(r==SSH_PACKET_USED)
-			break;
-	}
-	if(r==SSH_PACKET_NOT_USED){
-		SSH_LOG(SSH_LOG_RARE,"Couldn't do anything with packet type %d",type);
-		ssh_packet_send_unimplemented(session, session->recv_seq-1);
-	}
+    i = ssh_list_get_iterator(session->packet_callbacks);
+    while (i != NULL) {
+        cb = ssh_iterator_value(ssh_packet_callbacks, i);
+        i = i->next;
+
+        if (!cb) {
+            continue;
+        }
+
+        if (cb->start > type) {
+            continue;
+        }
+
+        if (cb->start + cb->n_callbacks <= type) {
+            continue;
+        }
+
+        if (cb->callbacks[type - cb->start] == NULL) {
+            continue;
+        }
+
+        rc = cb->callbacks[type - cb->start](session, type, session->in_buffer,
+                                             cb->user);
+        if (rc == SSH_PACKET_USED) {
+            break;
+        }
+    }
+
+    if (rc == SSH_PACKET_NOT_USED) {
+        SSH_LOG(SSH_LOG_RARE, "Couldn't do anything with packet type %d", type);
+        rc = ssh_packet_send_unimplemented(session, session->recv_seq - 1);
+        if (rc != SSH_OK) {
+            SSH_LOG(SSH_LOG_RARE, "Failed to send unimplemented: %s",
+                    ssh_get_error(session));
+        }
+    }
 }
 
 /** @internal
@@ -1361,109 +1570,392 @@ static int ssh_packet_write(ssh_session session) {
   return rc;
 }
 
-static int packet_send2(ssh_session session) {
-  unsigned int blocksize = (session->current_crypto ?
-      session->current_crypto->out_cipher->blocksize : 8);
-  unsigned int lenfield_blocksize = (session->current_crypto ?
-      session->current_crypto->out_cipher->lenfield_blocksize : 0);
-  enum ssh_hmac_e hmac_type = (session->current_crypto ?
-      session->current_crypto->out_hmac : session->next_crypto->out_hmac);
-  uint32_t currentlen = ssh_buffer_get_len(session->out_buffer);
-  unsigned char *hmac = NULL;
-  char padstring[32] = { 0 };
-  int rc = SSH_ERROR;
-  uint32_t finallen,payloadsize,compsize;
-  uint8_t padding;
-  ssh_buffer header_buffer = ssh_buffer_new();
+static int packet_send2(ssh_session session)
+{
+    unsigned int blocksize = 8;
+    unsigned int lenfield_blocksize = 0;
+    enum ssh_hmac_e hmac_type;
+    uint32_t currentlen = ssh_buffer_get_len(session->out_buffer);
+    struct ssh_crypto_struct *crypto = NULL;
+    unsigned char *hmac = NULL;
+    uint8_t padding_data[32] = { 0 };
+    uint8_t padding_size;
+    uint32_t finallen, payloadsize, compsize;
+    uint8_t header[5] = {0};
+    uint8_t type, *payload;
+    int rc = SSH_ERROR;
+    bool etm = false;
+    int etm_packet_offset = 0;
+
+    crypto = ssh_packet_get_current_crypto(session, SSH_DIRECTION_OUT);
+    if (crypto) {
+        blocksize = crypto->out_cipher->blocksize;
+        lenfield_blocksize = crypto->out_cipher->lenfield_blocksize;
+        hmac_type = crypto->out_hmac;
+        etm = crypto->out_hmac_etm;
+    } else {
+        hmac_type = session->next_crypto->out_hmac;
+    }
+
+    payload = (uint8_t *)ssh_buffer_get(session->out_buffer);
+    type = payload[0]; /* type is the first byte of the packet now */
+
+    payloadsize = currentlen;
+    if (etm) {
+        etm_packet_offset = sizeof(uint32_t);
+        lenfield_blocksize = 0;
+    }
 
-  payloadsize = currentlen;
 #ifdef WITH_ZLIB
-  if (session->current_crypto
-      && session->current_crypto->do_compress_out
-      && ssh_buffer_get_len(session->out_buffer)) {
-    if (compress_buffer(session,session->out_buffer) < 0) {
-      goto error;
+    if (crypto != NULL && crypto->do_compress_out &&
+        ssh_buffer_get_len(session->out_buffer) > 0) {
+        rc = compress_buffer(session,session->out_buffer);
+        if (rc < 0) {
+            goto error;
+        }
+        currentlen = ssh_buffer_get_len(session->out_buffer);
     }
-    currentlen = ssh_buffer_get_len(session->out_buffer);
-  }
 #endif /* WITH_ZLIB */
-  compsize = currentlen;
-  /* compressed payload + packet len (4) + padding len (1) */
-  /* totallen - lenfield_blocksize must be equal to 0 (mod blocksize) */
-  padding = (blocksize - ((blocksize - lenfield_blocksize + currentlen + 5) % blocksize));
-  if(padding < 4) {
-    padding += blocksize;
-  }
-
-  if (session->current_crypto != NULL) {
-      int ok;
-
-      ok = ssh_get_random(padstring, padding, 0);
-      if (!ok) {
-          ssh_set_error(session, SSH_FATAL, "PRNG error");
-          goto error;
-      }
-  }
-
-  if (header_buffer == NULL){
-    ssh_set_error_oom(session);
-    goto error;
-  }
-  finallen = currentlen + padding + 1;
-  rc = ssh_buffer_pack(header_buffer, "db", finallen, padding);
-  if (rc == SSH_ERROR){
-    goto error;
-  }
-
-  rc = ssh_buffer_prepend_data(session->out_buffer,
-                               ssh_buffer_get(header_buffer),
-                               ssh_buffer_get_len(header_buffer));
-  if (rc < 0) {
-    goto error;
-  }
-  rc = ssh_buffer_add_data(session->out_buffer, padstring, padding);
-  if (rc < 0) {
-    goto error;
-  }
-#ifdef WITH_PCAP
-  if (session->pcap_ctx) {
-      ssh_pcap_context_write(session->pcap_ctx,
-                             SSH_PCAP_DIR_OUT,
-                             ssh_buffer_get(session->out_buffer),
-                             ssh_buffer_get_len(session->out_buffer),
-                             ssh_buffer_get_len(session->out_buffer));
-  }
-#endif
-  hmac = ssh_packet_encrypt(session, ssh_buffer_get(session->out_buffer),
-      ssh_buffer_get_len(session->out_buffer));
-  if (hmac) {
-    rc = ssh_buffer_add_data(session->out_buffer, hmac, hmac_digest_len(hmac_type));
-    if (rc < 0) {
-      goto error;
+    compsize = currentlen;
+    /* compressed payload + packet len (4) + padding_size len (1) */
+    /* totallen - lenfield_blocksize - etm_packet_offset must be equal to 0 (mod blocksize) */
+    padding_size = (blocksize - ((blocksize - lenfield_blocksize - etm_packet_offset + currentlen + 5) % blocksize));
+    if (padding_size < 4) {
+        padding_size += blocksize;
     }
-  }
 
-  rc = ssh_packet_write(session);
-  session->send_seq++;
-  if (session->raw_counter != NULL) {
-      session->raw_counter->out_bytes += payloadsize;
-      session->raw_counter->out_packets++;
-  }
+    if (crypto != NULL) {
+        int ok;
 
-  SSH_LOG(SSH_LOG_PACKET,
-          "packet: wrote [len=%d,padding=%hhd,comp=%d,payload=%d]",
-          finallen, padding, compsize, payloadsize);
-  if (ssh_buffer_reinit(session->out_buffer) < 0) {
-    rc = SSH_ERROR;
-  }
+        ok = ssh_get_random(padding_data, padding_size, 0);
+        if (!ok) {
+            ssh_set_error(session, SSH_FATAL, "PRNG error");
+            goto error;
+        }
+    }
+
+    finallen = currentlen - etm_packet_offset + padding_size + 1;
+
+    PUSH_BE_U32(header, 0, finallen);
+    PUSH_BE_U8(header, 4, padding_size);
+
+    rc = ssh_buffer_prepend_data(session->out_buffer,
+                                 header,
+                                 sizeof(header));
+    if (rc < 0) {
+        goto error;
+    }
+
+    rc = ssh_buffer_add_data(session->out_buffer, padding_data, padding_size);
+    if (rc < 0) {
+        goto error;
+    }
+
+#ifdef WITH_PCAP
+    if (session->pcap_ctx != NULL) {
+        ssh_pcap_context_write(session->pcap_ctx,
+                               SSH_PCAP_DIR_OUT,
+                               ssh_buffer_get(session->out_buffer),
+                               ssh_buffer_get_len(session->out_buffer),
+                               ssh_buffer_get_len(session->out_buffer));
+    }
+#endif
+
+    hmac = ssh_packet_encrypt(session,
+                              ssh_buffer_get(session->out_buffer),
+                              ssh_buffer_get_len(session->out_buffer));
+    if (hmac != NULL) {
+        rc = ssh_buffer_add_data(session->out_buffer,
+                                 hmac,
+                                 hmac_digest_len(hmac_type));
+        if (rc < 0) {
+            goto error;
+        }
+    }
+
+    rc = ssh_packet_write(session);
+    if (rc == SSH_ERROR) {
+        goto error;
+    }
+    session->send_seq++;
+    if (crypto != NULL) {
+        struct ssh_cipher_struct *cipher = NULL;
+
+        cipher = crypto->out_cipher;
+        cipher->packets++;
+        cipher->blocks += payloadsize / cipher->blocksize;
+    }
+    if (session->raw_counter != NULL) {
+        session->raw_counter->out_bytes += payloadsize;
+        session->raw_counter->out_packets++;
+    }
+
+    SSH_LOG(SSH_LOG_PACKET,
+            "packet: wrote [type=%u, len=%u, padding_size=%hhd, comp=%u, "
+            "payload=%u]",
+            type,
+            finallen,
+            padding_size,
+            compsize,
+            payloadsize);
+
+    rc = ssh_buffer_reinit(session->out_buffer);
+    if (rc < 0) {
+        rc = SSH_ERROR;
+        goto error;
+    }
+
+    /* We sent the NEWKEYS so any further packet needs to be encrypted
+     * with the new keys. We can not switch both directions (need to decrypt
+     * peer NEWKEYS) and we do not want to wait for the peer NEWKEYS
+     * too, so we will switch only the OUT direction now.
+     */
+    if (type == SSH2_MSG_NEWKEYS) {
+        rc = ssh_packet_set_newkeys(session, SSH_DIRECTION_OUT);
+    }
 error:
-  if (header_buffer != NULL) {
-      ssh_buffer_free(header_buffer);
-  }
-  return rc; /* SSH_OK, AGAIN or ERROR */
+    return rc; /* SSH_OK, AGAIN or ERROR */
 }
 
+static bool
+ssh_packet_is_kex(unsigned char type)
+{
+    return type >= SSH2_MSG_DISCONNECT &&
+           type <= SSH2_MSG_KEX_DH_GEX_REQUEST &&
+           type != SSH2_MSG_SERVICE_REQUEST &&
+           type != SSH2_MSG_SERVICE_ACCEPT &&
+           type != SSH2_MSG_IGNORE &&
+           type != SSH2_MSG_EXT_INFO;
+}
 
-int ssh_packet_send(ssh_session session) {
-    return packet_send2(session);
+static bool
+ssh_packet_in_rekey(ssh_session session)
+{
+    /* We know we are rekeying if we are authenticated and the DH
+     * status is not finished
+     */
+    return (session->flags & SSH_SESSION_FLAG_AUTHENTICATED) &&
+           (session->dh_handshake_state != DH_STATE_FINISHED);
+}
+
+int ssh_packet_send(ssh_session session)
+{
+    uint32_t payloadsize;
+    uint8_t type, *payload;
+    bool need_rekey, in_rekey;
+    int rc;
+
+    payloadsize = ssh_buffer_get_len(session->out_buffer);
+    if (payloadsize < 1) {
+        return SSH_ERROR;
+    }
+
+    payload = (uint8_t *)ssh_buffer_get(session->out_buffer);
+    type = payload[0]; /* type is the first byte of the packet now */
+    need_rekey = ssh_packet_need_rekey(session, payloadsize);
+    in_rekey = ssh_packet_in_rekey(session);
+
+    /* The rekey is triggered here. After that, only the key exchange
+     * packets can be sent, until we send our NEWKEYS.
+     */
+    if (need_rekey || (in_rekey && !ssh_packet_is_kex(type))) {
+        if (need_rekey) {
+            SSH_LOG(SSH_LOG_PACKET, "Outgoing packet triggered rekey");
+        }
+        /* Queue the current packet -- we will send it after the rekey */
+        SSH_LOG(SSH_LOG_PACKET, "Queuing packet type %d", type);
+        rc = ssh_list_append(session->out_queue, session->out_buffer);
+        if (rc != SSH_OK) {
+            return SSH_ERROR;
+        }
+        session->out_buffer = ssh_buffer_new();
+        if (session->out_buffer == NULL) {
+            ssh_set_error_oom(session);
+            return SSH_ERROR;
+        }
+
+        if (need_rekey) {
+            /* Send the KEXINIT packet instead.
+             * This recursivelly calls the packet_send(), but it should
+             * not get into rekeying again.
+             * After that we need to handle the key exchange responses
+             * up to the point where we can send the rest of the queue.
+             */
+            return ssh_send_rekex(session);
+        }
+        return SSH_OK;
+    }
+
+    /* Send the packet normally */
+    rc = packet_send2(session);
+
+    /* We finished the key exchange so we can try to send our queue now */
+    if (rc == SSH_OK && type == SSH2_MSG_NEWKEYS) {
+        struct ssh_iterator *it;
+
+        for (it = ssh_list_get_iterator(session->out_queue);
+             it != NULL;
+             it = ssh_list_get_iterator(session->out_queue)) {
+            struct ssh_buffer_struct *next_buffer = NULL;
+
+            /* Peek only -- do not remove from queue yet */
+            next_buffer = (struct ssh_buffer_struct *)it->data;
+            payloadsize = ssh_buffer_get_len(next_buffer);
+            if (ssh_packet_need_rekey(session, payloadsize)) {
+                /* Sigh ... we still can not send this packet. Repeat. */
+                SSH_LOG(SSH_LOG_PACKET, "Queued packet triggered rekey");
+                return ssh_send_rekex(session);
+            }
+            ssh_buffer_free(session->out_buffer);
+            session->out_buffer = ssh_list_pop_head(struct ssh_buffer_struct *,
+                                                    session->out_queue);
+            payload = (uint8_t *)ssh_buffer_get(session->out_buffer);
+            type = payload[0];
+            SSH_LOG(SSH_LOG_PACKET, "Dequeue packet type %d", type);
+            rc = packet_send2(session);
+            if (rc != SSH_OK) {
+                return rc;
+            }
+        }
+    }
+
+    return rc;
+}
+
+static void
+ssh_init_rekey_state(struct ssh_session_struct *session,
+                     struct ssh_cipher_struct *cipher)
+{
+    /* Reset the counters: should be NOOP */
+    cipher->packets = 0;
+    cipher->blocks = 0;
+
+    /* Default rekey limits for ciphers as specified in RFC4344, Section 3.2 */
+    if (cipher->blocksize >= 16) {
+        /* For larger block size (L bits) use maximum of 2**(L/4) blocks */
+        cipher->max_blocks = (uint64_t)1 << (cipher->blocksize*2);
+    } else {
+        /* For smaller blocks use limit of 1 GB as recommended in RFC4253 */
+        cipher->max_blocks = ((uint64_t)1 << 30) / cipher->blocksize;
+    }
+    /* If we have limit provided by user, use the smaller one */
+    if (session->opts.rekey_data != 0) {
+        cipher->max_blocks = MIN(cipher->max_blocks,
+                                 session->opts.rekey_data / cipher->blocksize);
+    }
+
+    SSH_LOG(SSH_LOG_PROTOCOL,
+            "Set rekey after %" PRIu64 " blocks",
+            cipher->max_blocks);
+}
+
+/*
+ * Once we got SSH2_MSG_NEWKEYS we can switch next_crypto and
+ * current_crypto for our desired direction
+ */
+int
+ssh_packet_set_newkeys(ssh_session session,
+                       enum ssh_crypto_direction_e direction)
+{
+    int rc;
+
+    SSH_LOG(SSH_LOG_TRACE,
+            "called, direction =%s%s",
+            direction & SSH_DIRECTION_IN ? " IN " : "",
+            direction & SSH_DIRECTION_OUT ? " OUT " : "");
+
+    session->next_crypto->used |= direction;
+    if (session->current_crypto != NULL) {
+        if (session->current_crypto->used & direction) {
+            SSH_LOG(SSH_LOG_WARNING, "This direction isn't used anymore.");
+        }
+        /* Mark the current requested direction unused */
+        session->current_crypto->used &= ~direction;
+    }
+
+    /* Both sides switched: do the actual switch now */
+    if (session->next_crypto->used == SSH_DIRECTION_BOTH) {
+        size_t digest_len;
+
+        if (session->current_crypto != NULL) {
+            crypto_free(session->current_crypto);
+            session->current_crypto = NULL;
+        }
+
+        session->current_crypto = session->next_crypto;
+        session->current_crypto->used = SSH_DIRECTION_BOTH;
+
+        /* Initialize the next_crypto structure */
+        session->next_crypto = crypto_new();
+        if (session->next_crypto == NULL) {
+            ssh_set_error_oom(session);
+            return SSH_ERROR;
+        }
+
+        digest_len = session->current_crypto->digest_len;
+        session->next_crypto->session_id = malloc(digest_len);
+        if (session->next_crypto->session_id == NULL) {
+            ssh_set_error_oom(session);
+            return SSH_ERROR;
+        }
+
+        memcpy(session->next_crypto->session_id,
+               session->current_crypto->session_id,
+               digest_len);
+
+        return SSH_OK;
+    }
+
+    /* Initialize common structures so the next context can be used in
+     * either direction */
+    if (session->client) {
+        /* The server has this part already done */
+        rc = ssh_make_sessionid(session);
+        if (rc != SSH_OK) {
+            return SSH_ERROR;
+        }
+
+        /*
+         * Set the cryptographic functions for the next crypto
+         * (it is needed for ssh_generate_session_keys for key lengths)
+         */
+        rc = crypt_set_algorithms_client(session);
+        if (rc < 0) {
+            return SSH_ERROR;
+        }
+    }
+
+    if (ssh_generate_session_keys(session) < 0) {
+        return SSH_ERROR;
+    }
+
+    /* Initialize rekeying states */
+    ssh_init_rekey_state(session,
+                         session->next_crypto->out_cipher);
+    ssh_init_rekey_state(session,
+                         session->next_crypto->in_cipher);
+    if (session->opts.rekey_time != 0) {
+        ssh_timestamp_init(&session->last_rekey_time);
+        SSH_LOG(SSH_LOG_PROTOCOL, "Set rekey after %" PRIu32 " seconds",
+                session->opts.rekey_time/1000);
+    }
+
+    /* Initialize the encryption and decryption keys in next_crypto */
+    rc = session->next_crypto->in_cipher->set_decrypt_key(
+        session->next_crypto->in_cipher,
+        session->next_crypto->decryptkey,
+        session->next_crypto->decryptIV);
+    if (rc < 0) {
+        return SSH_ERROR;
+    }
+
+    rc = session->next_crypto->out_cipher->set_encrypt_key(
+        session->next_crypto->out_cipher,
+        session->next_crypto->encryptkey,
+        session->next_crypto->encryptIV);
+    if (rc < 0) {
+        return SSH_ERROR;
+    }
+
+    return SSH_OK;
 }
