@@ -59,14 +59,13 @@
 #include "libssh/messages.h"
 #include "libssh/options.h"
 #include "libssh/curve25519.h"
+#include "libssh/token.h"
 
 #define set_status(session, status) do {\
         if (session->common.callbacks && session->common.callbacks->connect_status_function) \
             session->common.callbacks->connect_status_function(session->common.callbacks->userdata, status); \
     } while (0)
 
-static int dh_handshake_server(ssh_session session);
-
 /**
  * @addtogroup libssh_server
  *
@@ -82,83 +81,110 @@ static int dh_handshake_server(ssh_session session);
  * options that are currently set in the given ssh_session structure.
  */
 
-static int server_set_kex(ssh_session session) {
-  struct ssh_kex_struct *server = &session->next_crypto->server_kex;
-  int i, j, rc;
-  const char *wanted;
-  char hostkeys[128] = {0};
-  enum ssh_keytypes_e keytype;
-  size_t len;
-  int ok;
+int server_set_kex(ssh_session session)
+{
+    struct ssh_kex_struct *server = &session->next_crypto->server_kex;
+    int i, j, rc;
+    const char *wanted, *allowed;
+    char *kept;
+    char hostkeys[128] = {0};
+    enum ssh_keytypes_e keytype;
+    size_t len;
+    int ok;
 
-  ZERO_STRUCTP(server);
+    ZERO_STRUCTP(server);
 
-  ok = ssh_get_random(server->cookie, 16, 0);
-  if (!ok) {
-      ssh_set_error(session, SSH_FATAL, "PRNG error");
-      return -1;
-  }
+    ok = ssh_get_random(server->cookie, 16, 0);
+    if (!ok) {
+        ssh_set_error(session, SSH_FATAL, "PRNG error");
+        return -1;
+    }
 
-  if (session->srv.ed25519_key != NULL) {
-      snprintf(hostkeys,
-               sizeof(hostkeys),
-               "%s",
-               ssh_key_type_to_char(ssh_key_type(session->srv.ed25519_key)));
-  }
+    if (session->srv.ed25519_key != NULL) {
+        snprintf(hostkeys,
+                 sizeof(hostkeys),
+                 "%s",
+                 ssh_key_type_to_char(ssh_key_type(session->srv.ed25519_key)));
+    }
 #ifdef HAVE_ECC
-  if (session->srv.ecdsa_key != NULL) {
-	  len = strlen(hostkeys);
-      snprintf(hostkeys + len, sizeof(hostkeys) - len,
-               ",%s", session->srv.ecdsa_key->type_c);
-  }
+    if (session->srv.ecdsa_key != NULL) {
+        len = strlen(hostkeys);
+        snprintf(hostkeys + len, sizeof(hostkeys) - len,
+                 ",%s", session->srv.ecdsa_key->type_c);
+    }
 #endif
 #ifdef HAVE_DSA
-  if (session->srv.dsa_key != NULL) {
-      len = strlen(hostkeys);
-      keytype = ssh_key_type(session->srv.dsa_key);
+    if (session->srv.dsa_key != NULL) {
+        len = strlen(hostkeys);
+        keytype = ssh_key_type(session->srv.dsa_key);
 
-      snprintf(hostkeys + len, sizeof(hostkeys) - len,
-               ",%s", ssh_key_type_to_char(keytype));
-  }
+        snprintf(hostkeys + len, sizeof(hostkeys) - len,
+                 ",%s", ssh_key_type_to_char(keytype));
+    }
 #endif
-  if (session->srv.rsa_key != NULL) {
-      /* We support also the SHA2 variants */
-      len = strlen(hostkeys);
-      snprintf(hostkeys + len, sizeof(hostkeys) - len,
-               ",rsa-sha2-512,rsa-sha2-256");
+    if (session->srv.rsa_key != NULL) {
+        /* We support also the SHA2 variants */
+        len = strlen(hostkeys);
+        snprintf(hostkeys + len, sizeof(hostkeys) - len,
+                 ",rsa-sha2-512,rsa-sha2-256");
 
-      len = strlen(hostkeys);
-      keytype = ssh_key_type(session->srv.rsa_key);
+        len = strlen(hostkeys);
+        keytype = ssh_key_type(session->srv.rsa_key);
 
-      snprintf(hostkeys + len, sizeof(hostkeys) - len,
-               ",%s", ssh_key_type_to_char(keytype));
-  }
-
-  if (strlen(hostkeys) == 0) {
-      return -1;
-  }
-
-  rc = ssh_options_set_algo(session,
-                            SSH_HOSTKEYS,
-                            hostkeys[0] == ',' ? hostkeys + 1 : hostkeys);
-  if (rc < 0) {
-      return -1;
-  }
-
-  for (i = 0; i < 10; i++) {
-    if ((wanted = session->opts.wanted_methods[i]) == NULL) {
-      wanted = ssh_kex_get_supported_method(i);
+        snprintf(hostkeys + len, sizeof(hostkeys) - len,
+                 ",%s", ssh_key_type_to_char(keytype));
     }
-    server->methods[i] = strdup(wanted);
-    if (server->methods[i] == NULL) {
-      for (j = 0; j < i; j++) {
-        SAFE_FREE(server->methods[j]);
-      }
-      return -1;
-    }
-  }
 
-  return 0;
+    if (strlen(hostkeys) == 0) {
+        return -1;
+    }
+
+    if (session->opts.wanted_methods[SSH_HOSTKEYS]) {
+        allowed = session->opts.wanted_methods[SSH_HOSTKEYS];
+    } else {
+        if (ssh_fips_mode()) {
+            allowed = ssh_kex_get_fips_methods(SSH_HOSTKEYS);
+        } else {
+            allowed = ssh_kex_get_default_methods(SSH_HOSTKEYS);
+        }
+    }
+
+    /* It is expected for the list of allowed hostkeys to be ordered by
+     * preference */
+    kept = ssh_find_all_matching(hostkeys[0] == ',' ? hostkeys + 1 : hostkeys,
+                                 allowed);
+    if (kept == NULL) {
+        /* Nothing was allowed */
+        return -1;
+    }
+
+    rc = ssh_options_set_algo(session,
+                              SSH_HOSTKEYS,
+                              kept);
+    SAFE_FREE(kept);
+    if (rc < 0) {
+        return -1;
+    }
+
+    for (i = 0; i < 10; i++) {
+        wanted = session->opts.wanted_methods[i];
+        if (wanted  == NULL) {
+            if (ssh_fips_mode()) {
+                wanted = ssh_kex_get_fips_methods(i);
+            } else {
+                wanted = ssh_kex_get_default_methods(i);
+            }
+        }
+        server->methods[i] = strdup(wanted);
+        if (server->methods[i] == NULL) {
+            for (j = 0; j < i; j++) {
+                SAFE_FREE(server->methods[j]);
+            }
+            return -1;
+        }
+    }
+
+    return 0;
 }
 
 int ssh_server_init_kex(ssh_session session) {
@@ -176,38 +202,23 @@ int ssh_server_init_kex(ssh_session session) {
     return server_set_kex(session);
 }
 
-/** @internal
- * @brief parse an incoming SSH_MSG_KEXDH_INIT packet and complete
- *        key exchange
- **/
-static int ssh_server_kexdh_init(ssh_session session, ssh_buffer packet){
-    ssh_string e;
-    e = ssh_buffer_get_ssh_string(packet);
-    if (e == NULL) {
-      ssh_set_error(session, SSH_FATAL, "No e number in client request");
-      return -1;
-    }
-    if (ssh_dh_import_e(session, e) < 0) {
-      ssh_set_error(session, SSH_FATAL, "Cannot import e number");
-      session->session_state=SSH_SESSION_STATE_ERROR;
-    } else {
-      session->dh_handshake_state=DH_STATE_INIT_SENT;
-      dh_handshake_server(session);
-    }
-    ssh_string_free(e);
-    return SSH_OK;
-}
-
 static int ssh_server_send_extensions(ssh_session session) {
     int rc;
     const char *hostkey_algorithms;
 
     SSH_LOG(SSH_LOG_PACKET, "Sending SSH_MSG_EXT_INFO");
-    /*
-     * We can list here all the default hostkey methods, since
-     * they already contain the SHA2 extension algorithms
-     */
-    hostkey_algorithms = ssh_kex_get_default_methods(SSH_HOSTKEYS);
+
+    if (session->opts.pubkey_accepted_types) {
+        hostkey_algorithms = session->opts.pubkey_accepted_types;
+    } else {
+        if (ssh_fips_mode()) {
+            hostkey_algorithms = ssh_kex_get_fips_methods(SSH_HOSTKEYS);
+        } else {
+            /* There are no restrictions to the accepted public keys */
+            hostkey_algorithms = ssh_kex_get_default_methods(SSH_HOSTKEYS);
+        }
+    }
+
     rc = ssh_buffer_pack(session->out_buffer,
                          "bdss",
                          SSH2_MSG_EXT_INFO,
@@ -230,14 +241,15 @@ error:
 }
 
 SSH_PACKET_CALLBACK(ssh_packet_kexdh_init){
-  int rc = SSH_ERROR;
+  (void)packet;
   (void)type;
   (void)user;
 
   SSH_LOG(SSH_LOG_PACKET,"Received SSH_MSG_KEXDH_INIT");
   if(session->dh_handshake_state != DH_STATE_INIT){
     SSH_LOG(SSH_LOG_RARE,"Invalid state for SSH_MSG_KEXDH_INIT");
-    goto error;
+    session->session_state = SSH_SESSION_STATE_ERROR;
+    return SSH_PACKET_USED;
   }
 
   /* If first_kex_packet_follows guess was wrong, ignore this message. */
@@ -245,41 +257,11 @@ SSH_PACKET_CALLBACK(ssh_packet_kexdh_init){
     SSH_LOG(SSH_LOG_RARE, "first_kex_packet_follows guess was wrong, "
                           "ignoring first SSH_MSG_KEXDH_INIT message");
     session->first_kex_follows_guess_wrong = 0;
-    rc = SSH_OK;
-    goto error;
-  }
 
-  switch(session->next_crypto->kex_type){
-      case SSH_KEX_DH_GROUP1_SHA1:
-      case SSH_KEX_DH_GROUP14_SHA1:
-      case SSH_KEX_DH_GROUP16_SHA512:
-      case SSH_KEX_DH_GROUP18_SHA512:
-        rc=ssh_server_kexdh_init(session, packet);
-        break;
-  #ifdef HAVE_ECDH
-      case SSH_KEX_ECDH_SHA2_NISTP256:
-      case SSH_KEX_ECDH_SHA2_NISTP384:
-      case SSH_KEX_ECDH_SHA2_NISTP521:
-        rc = ssh_server_ecdh_init(session, packet);
-        break;
-  #endif
-  #ifdef HAVE_CURVE25519
-      case SSH_KEX_CURVE25519_SHA256:
-      case SSH_KEX_CURVE25519_SHA256_LIBSSH_ORG:
-    	rc = ssh_server_curve25519_init(session, packet);
-    	break;
-  #endif
-      default:
-        ssh_set_error(session,SSH_FATAL,"Wrong kex type in ssh_packet_kexdh_init");
-        rc = SSH_ERROR;
+    return SSH_PACKET_USED;
   }
-
-error:
-  if (rc == SSH_ERROR) {
-      session->session_state = SSH_SESSION_STATE_ERROR;
-  }
-
-  return SSH_PACKET_USED;
+  SSH_LOG(SSH_LOG_DEBUG, "Calling next KEXDH handler");
+  return SSH_PACKET_NOT_USED;
 }
 
 int ssh_get_key_params(ssh_session session, ssh_key *privkey){
@@ -294,7 +276,9 @@ int ssh_get_key_params(ssh_session session, ssh_key *privkey){
       case SSH_KEYTYPE_RSA:
         *privkey = session->srv.rsa_key;
         break;
-      case SSH_KEYTYPE_ECDSA:
+      case SSH_KEYTYPE_ECDSA_P256:
+      case SSH_KEYTYPE_ECDSA_P384:
+      case SSH_KEYTYPE_ECDSA_P521:
         *privkey = session->srv.ecdsa_key;
         break;
       case SSH_KEYTYPE_ED25519:
@@ -333,93 +317,6 @@ int ssh_get_key_params(ssh_session session, ssh_key *privkey){
     return SSH_OK;
 }
 
-static int dh_handshake_server(ssh_session session) {
-  ssh_key privkey;
-  ssh_string sig_blob;
-  ssh_string f;
-  ssh_string pubkey_blob = NULL;
-  int rc;
-
-  if (ssh_dh_generate_y(session) < 0) {
-    ssh_set_error(session, SSH_FATAL, "Could not create y number");
-    return -1;
-  }
-  if (ssh_dh_generate_f(session) < 0) {
-    ssh_set_error(session, SSH_FATAL, "Could not create f number");
-    return -1;
-  }
-
-  f = ssh_dh_get_f(session);
-  if (f == NULL) {
-    ssh_set_error(session, SSH_FATAL, "Could not get the f number");
-    return -1;
-  }
-
-  if (ssh_get_key_params(session,&privkey) != SSH_OK){
-      ssh_string_free(f);
-      return -1;
-  }
-
-  if (ssh_dh_build_k(session) < 0) {
-    ssh_set_error(session, SSH_FATAL, "Could not import the public key");
-    ssh_string_free(f);
-    return -1;
-  }
-
-  if (ssh_make_sessionid(session) != SSH_OK) {
-    ssh_set_error(session, SSH_FATAL, "Could not create a session id");
-    ssh_string_free(f);
-    return -1;
-  }
-
-  sig_blob = ssh_srv_pki_do_sign_sessionid(session, privkey);
-  if (sig_blob == NULL) {
-    ssh_set_error(session, SSH_FATAL, "Could not sign the session id");
-    ssh_string_free(f);
-    return -1;
-  }
-
-  rc = ssh_dh_get_next_server_publickey_blob(session, &pubkey_blob);
-  if (rc != SSH_OK) {
-      ssh_set_error_oom(session);
-      ssh_string_free(f);
-      ssh_string_free(sig_blob);
-      return -1;
-  }
-
-  rc = ssh_buffer_pack(session->out_buffer,
-                       "bSSS",
-                       SSH2_MSG_KEXDH_REPLY,
-                       pubkey_blob,
-                       f,
-                       sig_blob);
-  ssh_string_free(f);
-  ssh_string_free(sig_blob);
-  ssh_string_free(pubkey_blob);
-  if(rc != SSH_OK){
-    ssh_set_error_oom(session);
-    ssh_buffer_reinit(session->out_buffer);
-    return -1;
-  }
-
-  if (ssh_packet_send(session) == SSH_ERROR) {
-    return -1;
-  }
-
-  if (ssh_buffer_add_u8(session->out_buffer, SSH2_MSG_NEWKEYS) < 0) {
-    ssh_buffer_reinit(session->out_buffer);
-    return -1;
-  }
-
-  if (ssh_packet_send(session) == SSH_ERROR) {
-    return -1;
-  }
-  SSH_LOG(SSH_LOG_PACKET, "SSH_MSG_NEWKEYS sent");
-  session->dh_handshake_state=DH_STATE_NEWKEYS_SENT;
-
-  return 0;
-}
-
 /**
  * @internal
  *
@@ -439,7 +336,7 @@ static void ssh_server_connection_callback(ssh_session session){
                 goto error;
             }
             set_status(session, 0.4f);
-            SSH_LOG(SSH_LOG_RARE,
+            SSH_LOG(SSH_LOG_PROTOCOL,
                     "SSH client banner: %s", session->clientbanner);
 
             /* Here we analyze the different protocols the server allows. */
@@ -485,37 +382,9 @@ static void ssh_server_connection_callback(ssh_session session){
             break;
         case SSH_SESSION_STATE_DH:
             if(session->dh_handshake_state==DH_STATE_FINISHED){
-                if (ssh_generate_session_keys(session) < 0) {
-                    goto error;
-                }
 
-                /*
-                 * Once we got SSH2_MSG_NEWKEYS we can switch next_crypto and
-                 * current_crypto
-                 */
-                if (session->current_crypto) {
-                    crypto_free(session->current_crypto);
-                }
-
-                /* FIXME TODO later, include a function to change keys */
-                session->current_crypto = session->next_crypto;
-                session->next_crypto = crypto_new();
-                if (session->next_crypto == NULL) {
-                    goto error;
-                }
-                session->next_crypto->session_id = malloc(session->current_crypto->digest_len);
-                if (session->next_crypto->session_id == NULL) {
-                    ssh_set_error_oom(session);
-                    goto error;
-                }
-                memcpy(session->next_crypto->session_id, session->current_crypto->session_id,
-                        session->current_crypto->digest_len);
-                if (session->current_crypto->in_cipher->set_decrypt_key(session->current_crypto->in_cipher, session->current_crypto->decryptkey,
-                            session->current_crypto->decryptIV) < 0) {
-                    goto error;
-                }
-                if (session->current_crypto->out_cipher->set_encrypt_key(session->current_crypto->out_cipher, session->current_crypto->encryptkey,
-                            session->current_crypto->encryptIV) < 0) {
+                rc = ssh_packet_set_newkeys(session, SSH_DIRECTION_IN);
+                if (rc != SSH_OK) {
                     goto error;
                 }
 
@@ -526,7 +395,22 @@ static void ssh_server_connection_callback(ssh_session session){
                  */
                 if (session->extensions & SSH_EXT_NEGOTIATION &&
                     session->session_state != SSH_SESSION_STATE_AUTHENTICATED) {
-                    ssh_server_send_extensions(session);
+
+                    /*
+                     * Only send an SSH_MSG_EXT_INFO message the first time the client
+                     * undergoes NEWKEYS.  It is unexpected for this message to be sent
+                     * upon rekey, and may cause clients to log error messages.
+                     *
+                     * The session_state can not be used for this purpose because it is
+                     * re-set to SSH_SESSION_STATE_KEXINIT_RECEIVED during rekey.  So,
+                     * use the connected flag which transitions from non-zero below.
+                     *
+                     * See also:
+                     * - https://bugzilla.mindrot.org/show_bug.cgi?id=2929
+                     */
+                    if (session->connected == 0) {
+                        ssh_server_send_extensions(session);
+                    }
                 }
 
                 set_status(session,1.0f);
@@ -1047,34 +931,47 @@ int ssh_message_auth_interactive_request(ssh_message msg, const char *name,
   return rc;
 }
 
-int ssh_auth_reply_success(ssh_session session, int partial) {
-  int r;
+int ssh_auth_reply_success(ssh_session session, int partial)
+{
+    struct ssh_crypto_struct *crypto = NULL;
+    int r;
 
-  if (session == NULL) {
-	  return SSH_ERROR;
-  }
+    if (session == NULL) {
+        return SSH_ERROR;
+    }
 
-  if (partial) {
-    return ssh_auth_reply_default(session, partial);
-  }
-  
-  session->session_state = SSH_SESSION_STATE_AUTHENTICATED;
-  session->flags |= SSH_SESSION_FLAG_AUTHENTICATED;
+    if (partial) {
+        return ssh_auth_reply_default(session, partial);
+    }
 
-  if (ssh_buffer_add_u8(session->out_buffer,SSH2_MSG_USERAUTH_SUCCESS) < 0) {
-    return SSH_ERROR;
-  }
+    r = ssh_buffer_add_u8(session->out_buffer,SSH2_MSG_USERAUTH_SUCCESS);
+    if (r < 0) {
+        return SSH_ERROR;
+    }
 
-  r = ssh_packet_send(session);
-  if(session->current_crypto && session->current_crypto->delayed_compress_out){
-      SSH_LOG(SSH_LOG_PROTOCOL,"Enabling delayed compression OUT");
-  	session->current_crypto->do_compress_out=1;
-  }
-  if(session->current_crypto && session->current_crypto->delayed_compress_in){
-      SSH_LOG(SSH_LOG_PROTOCOL,"Enabling delayed compression IN");
-  	session->current_crypto->do_compress_in=1;
-  }
-  return r;
+    r = ssh_packet_send(session);
+
+    /*
+     * Consider the session as having been authenticated only after sending
+     * the USERAUTH_SUCCESS message.  Setting these flags after ssh_packet_send
+     * ensures that a rekey is not triggered prematurely, causing the message
+     * to be queued.
+     */
+    session->session_state = SSH_SESSION_STATE_AUTHENTICATED;
+    session->flags |= SSH_SESSION_FLAG_AUTHENTICATED;
+
+    crypto = ssh_packet_get_current_crypto(session, SSH_DIRECTION_OUT);
+    if (crypto != NULL && crypto->delayed_compress_out) {
+        SSH_LOG(SSH_LOG_PROTOCOL, "Enabling delayed compression OUT");
+        crypto->do_compress_out = 1;
+    }
+
+    crypto = ssh_packet_get_current_crypto(session, SSH_DIRECTION_IN);
+    if (crypto != NULL && crypto->delayed_compress_in) {
+        SSH_LOG(SSH_LOG_PROTOCOL, "Enabling delayed compression IN");
+        crypto->do_compress_in = 1;
+    }
+    return r;
 }
 
 int ssh_message_auth_reply_success(ssh_message msg, int partial) {
