@@ -48,6 +48,7 @@
 #include "libssh/channels.h"
 #include "libssh/session.h"
 #include "libssh/misc.h"
+#include "libssh/bytearray.h"
 
 #ifdef WITH_SFTP
 
@@ -67,19 +68,6 @@ static void sftp_message_free(sftp_message msg);
 static void sftp_set_error(sftp_session sftp, int errnum);
 static void status_msg_free(sftp_status_message status);
 
-static uint32_t sftp_get_u32(const void *vp)
-{
-    const uint8_t *p = (const uint8_t *)vp;
-    uint32_t v;
-
-    v  = (uint32_t)p[0] << 24;
-    v |= (uint32_t)p[1] << 16;
-    v |= (uint32_t)p[2] << 8;
-    v |= (uint32_t)p[3];
-
-    return v;
-}
-
 static sftp_ext sftp_ext_new(void) {
   sftp_ext ext;
 
@@ -91,23 +79,31 @@ static sftp_ext sftp_ext_new(void) {
   return ext;
 }
 
-static void sftp_ext_free(sftp_ext ext) {
-  unsigned int i;
+static void sftp_ext_free(sftp_ext ext)
+{
+    size_t i;
 
-  if (ext == NULL) {
-    return;
-  }
-
-  if (ext->count) {
-    for (i = 0; i < ext->count; i++) {
-      SAFE_FREE(ext->name[i]);
-      SAFE_FREE(ext->data[i]);
+    if (ext == NULL) {
+        return;
     }
-    SAFE_FREE(ext->name);
-    SAFE_FREE(ext->data);
-  }
 
-  SAFE_FREE(ext);
+    if (ext->count > 0) {
+        if (ext->name != NULL) {
+            for (i = 0; i < ext->count; i++) {
+                SAFE_FREE(ext->name[i]);
+            }
+            SAFE_FREE(ext->name);
+        }
+
+        if (ext->data != NULL) {
+            for (i = 0; i < ext->count; i++) {
+                SAFE_FREE(ext->data[i]);
+            }
+            SAFE_FREE(ext->data);
+        }
+    }
+
+    SAFE_FREE(ext);
 }
 
 sftp_session sftp_new(ssh_session session)
@@ -127,22 +123,26 @@ sftp_session sftp_new(ssh_session session)
 
     sftp->ext = sftp_ext_new();
     if (sftp->ext == NULL) {
+        ssh_set_error_oom(session);
         goto error;
     }
 
     sftp->read_packet = calloc(1, sizeof(struct sftp_packet_struct));
     if (sftp->read_packet == NULL) {
+        ssh_set_error_oom(session);
         goto error;
     }
 
     sftp->read_packet->payload = ssh_buffer_new();
     if (sftp->read_packet->payload == NULL) {
+        ssh_set_error_oom(session);
         goto error;
     }
 
     sftp->session = session;
     sftp->channel = ssh_channel_new(session);
     if (sftp->channel == NULL) {
+        ssh_set_error_oom(session);
         goto error;
     }
 
@@ -156,7 +156,6 @@ sftp_session sftp_new(ssh_session session)
 
     return sftp;
 error:
-    ssh_set_error_oom(session);
     if (sftp->ext != NULL) {
         sftp_ext_free(sftp->ext);
     }
@@ -289,7 +288,7 @@ int sftp_server_init(sftp_session sftp){
   }
   ssh_buffer_free(reply);
 
-  SSH_LOG(SSH_LOG_RARE, "Server version sent");
+  SSH_LOG(SSH_LOG_PROTOCOL, "Server version sent");
 
   if (version > LIBSFTP_VERSION) {
     sftp->version = LIBSFTP_VERSION;
@@ -299,6 +298,32 @@ int sftp_server_init(sftp_session sftp){
 
   return 0;
 }
+
+void sftp_server_free(sftp_session sftp)
+{
+    sftp_request_queue ptr;
+
+    if (sftp == NULL) {
+        return;
+    }
+
+    ptr = sftp->queue;
+    while(ptr) {
+        sftp_request_queue old;
+        sftp_message_free(ptr->message);
+        old = ptr->next;
+        SAFE_FREE(ptr);
+        ptr = old;
+    }
+
+    SAFE_FREE(sftp->handles);
+    SSH_BUFFER_FREE(sftp->read_packet->payload);
+    SAFE_FREE(sftp->read_packet);
+
+    sftp_ext_free(sftp->ext);
+
+    SAFE_FREE(sftp);
+}
 #endif /* WITH_SERVER */
 
 void sftp_free(sftp_session sftp)
@@ -309,17 +334,20 @@ void sftp_free(sftp_session sftp)
         return;
     }
 
-    ssh_channel_send_eof(sftp->channel);
-    ptr = sftp->queue;
-    while(ptr) {
-        sftp_request_queue old;
-        sftp_message_free(ptr->message);
-        old = ptr->next;
-        SAFE_FREE(ptr);
-        ptr = old;
-    }
+    if (sftp->channel != NULL) {
+        ssh_channel_send_eof(sftp->channel);
+        ptr = sftp->queue;
+        while(ptr) {
+            sftp_request_queue old;
+            sftp_message_free(ptr->message);
+            old = ptr->next;
+            SAFE_FREE(ptr);
+            ptr = old;
+        }
 
-    ssh_channel_free(sftp->channel);
+        ssh_channel_free(sftp->channel);
+        sftp->channel = NULL;
+    }
 
     SAFE_FREE(sftp->handles);
     SSH_BUFFER_FREE(sftp->read_packet->payload);
@@ -330,32 +358,41 @@ void sftp_free(sftp_session sftp)
     SAFE_FREE(sftp);
 }
 
-int sftp_packet_write(sftp_session sftp, uint8_t type, ssh_buffer payload){
-  int size;
+int sftp_packet_write(sftp_session sftp, uint8_t type, ssh_buffer payload)
+{
+    uint8_t header[5] = {0};
+    uint32_t payload_size;
+    int size;
+    int rc;
 
-  if (ssh_buffer_prepend_data(payload, &type, sizeof(uint8_t)) < 0) {
-    ssh_set_error_oom(sftp->session);
-    return -1;
-  }
+    /* Add size of type */
+    payload_size = ssh_buffer_get_len(payload) + sizeof(uint8_t);
+    PUSH_BE_U32(header, 0, payload_size);
+    PUSH_BE_U8(header, 4, type);
 
-  size = htonl(ssh_buffer_get_len(payload));
-  if (ssh_buffer_prepend_data(payload, &size, sizeof(uint32_t)) < 0) {
-    ssh_set_error_oom(sftp->session);
-    return -1;
-  }
+    rc = ssh_buffer_prepend_data(payload, header, sizeof(header));
+    if (rc < 0) {
+        ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
+        return -1;
+    }
 
-  size = ssh_channel_write(sftp->channel, ssh_buffer_get(payload),
-      ssh_buffer_get_len(payload));
-  if (size < 0) {
-    return -1;
-  } else if((uint32_t) size != ssh_buffer_get_len(payload)) {
-    SSH_LOG(SSH_LOG_PACKET,
-        "Had to write %d bytes, wrote only %d",
-        ssh_buffer_get_len(payload),
-        size);
-  }
+    size = ssh_channel_write(sftp->channel,
+                             ssh_buffer_get(payload),
+                             ssh_buffer_get_len(payload));
+    if (size < 0) {
+        sftp_set_error(sftp, SSH_FX_FAILURE);
+        return -1;
+    }
 
-  return size;
+    if ((uint32_t)size != ssh_buffer_get_len(payload)) {
+        SSH_LOG(SSH_LOG_PACKET,
+                "Had to write %d bytes, wrote only %d",
+                ssh_buffer_get_len(payload),
+                size);
+    }
+
+    return size;
 }
 
 sftp_packet sftp_packet_read(sftp_session sftp)
@@ -377,12 +414,14 @@ sftp_packet sftp_packet_read(sftp_session sftp)
         rc = ssh_buffer_reinit(packet->payload);
         if (rc != 0) {
             ssh_set_error_oom(sftp->session);
+            sftp_set_error(sftp, SSH_FX_FAILURE);
             return NULL;
         }
     } else {
         packet->payload = ssh_buffer_new();
         if (packet->payload == NULL) {
             ssh_set_error_oom(sftp->session);
+            sftp_set_error(sftp, SSH_FX_FAILURE);
             return NULL;
         }
     }
@@ -398,6 +437,10 @@ sftp_packet sftp_packet_read(sftp_session sftp)
         } else if (s == 0) {
             is_eof = ssh_channel_is_eof(sftp->channel);
             if (is_eof) {
+                ssh_set_error(sftp->session,
+                              SSH_FATAL,
+                              "Received EOF while reading sftp packet size");
+                sftp_set_error(sftp, SSH_FX_EOF);
                 goto error;
             }
         } else {
@@ -405,9 +448,10 @@ sftp_packet sftp_packet_read(sftp_session sftp)
         }
     } while (nread < 4);
 
-    size = sftp_get_u32(buffer);
+    size = PULL_BE_U32(buffer, 0);
     if (size == 0 || size > SFTP_PACKET_SIZE_MAX) {
         ssh_set_error(sftp->session, SSH_FATAL, "Invalid sftp packet size!");
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         goto error;
     }
 
@@ -418,6 +462,10 @@ sftp_packet sftp_packet_read(sftp_session sftp)
         } else if (nread == 0) {
             is_eof = ssh_channel_is_eof(sftp->channel);
             if (is_eof) {
+                ssh_set_error(sftp->session,
+                              SSH_FATAL,
+                              "Received EOF while reading sftp packet type");
+                sftp_set_error(sftp, SSH_FX_EOF);
                 goto error;
             }
         }
@@ -431,6 +479,7 @@ sftp_packet sftp_packet_read(sftp_session sftp)
     nread = ssh_buffer_allocate_size(packet->payload, size);
     if (nread < 0) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         goto error;
     }
     while (size > 0 && size < SFTP_PACKET_SIZE_MAX) {
@@ -447,12 +496,17 @@ sftp_packet sftp_packet_read(sftp_session sftp)
             rc = ssh_buffer_add_data(packet->payload, buffer, nread);
             if (rc != 0) {
                 ssh_set_error_oom(sftp->session);
+                sftp_set_error(sftp, SSH_FX_FAILURE);
                 goto error;
             }
         } else { /* nread == 0 */
             /* Retry the reading unless the remote was closed */
             is_eof = ssh_channel_is_eof(sftp->channel);
             if (is_eof) {
+                ssh_set_error(sftp->session,
+                              SSH_REQUEST_DENIED,
+                              "Received EOF while reading sftp packet");
+                sftp_set_error(sftp, SSH_FX_EOF);
                 goto error;
             }
         }
@@ -510,12 +564,14 @@ static sftp_message sftp_get_message(sftp_packet packet)
                       SSH_FATAL,
                       "Unknown packet type %d",
                       packet->type);
+        sftp_set_error(packet->sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
     msg = calloc(1, sizeof(struct sftp_message_struct));
     if (msg == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(packet->sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -531,6 +587,7 @@ static sftp_message sftp_get_message(sftp_packet packet)
         ssh_set_error(packet->sftp->session, SSH_FATAL,
                 "Invalid packet %d: no ID", packet->type);
         sftp_message_free(msg);
+        sftp_set_error(packet->sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -587,6 +644,7 @@ int sftp_init(sftp_session sftp) {
   buffer = ssh_buffer_new();
   if (buffer == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -594,6 +652,7 @@ int sftp_init(sftp_session sftp) {
   if (rc != SSH_OK) {
     ssh_set_error_oom(sftp->session);
     ssh_buffer_free(buffer);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
   if (sftp_packet_write(sftp, SSH_FXP_INIT, buffer) < 0) {
@@ -616,9 +675,10 @@ int sftp_init(sftp_session sftp) {
   /* TODO: are we sure there are 4 bytes ready? */
   rc = ssh_buffer_unpack(packet->payload, "d", &version);
   if (rc != SSH_OK){
+      sftp_set_error(sftp, SSH_FX_FAILURE);
       return -1;
   }
-  SSH_LOG(SSH_LOG_RARE,
+  SSH_LOG(SSH_LOG_PROTOCOL,
       "SFTP server version %d",
       version);
   rc = ssh_buffer_unpack(packet->payload, "s", &ext_name);
@@ -631,7 +691,7 @@ int sftp_init(sftp_session sftp) {
       break;
     }
 
-    SSH_LOG(SSH_LOG_RARE,
+    SSH_LOG(SSH_LOG_PROTOCOL,
         "SFTP server extension: %s, version: %s",
         ext_name, ext_data);
 
@@ -641,6 +701,7 @@ int sftp_init(sftp_session sftp) {
       ssh_set_error_oom(sftp->session);
       SAFE_FREE(ext_name);
       SAFE_FREE(ext_data);
+      sftp_set_error(sftp, SSH_FX_FAILURE);
       return -1;
     }
     tmp[count - 1] = ext_name;
@@ -651,6 +712,7 @@ int sftp_init(sftp_session sftp) {
       ssh_set_error_oom(sftp->session);
       SAFE_FREE(ext_name);
       SAFE_FREE(ext_data);
+      sftp_set_error(sftp, SSH_FX_FAILURE);
       return -1;
     }
     tmp[count - 1] = ext_data;
@@ -736,6 +798,7 @@ static sftp_request_queue request_queue_new(sftp_message msg) {
   queue = calloc(1, sizeof(struct sftp_request_queue_struct));
   if (queue == NULL) {
     ssh_set_error_oom(msg->sftp->session);
+    sftp_set_error(msg->sftp, SSH_FX_FAILURE);
     return NULL;
   }
 
@@ -832,12 +895,14 @@ static sftp_status_message parse_status_msg(sftp_message msg){
   if (msg->packet_type != SSH_FXP_STATUS) {
     ssh_set_error(msg->sftp->session, SSH_FATAL,
         "Not a ssh_fxp_status message passed in!");
+    sftp_set_error(msg->sftp, SSH_FX_BAD_MESSAGE);
     return NULL;
   }
 
   status = calloc(1, sizeof(struct sftp_status_message_struct));
   if (status == NULL) {
     ssh_set_error_oom(msg->sftp->session);
+    sftp_set_error(msg->sftp, SSH_FX_FAILURE);
     return NULL;
   }
 
@@ -848,6 +913,7 @@ static sftp_status_message parse_status_msg(sftp_message msg){
     SAFE_FREE(status);
     ssh_set_error(msg->sftp->session, SSH_FATAL,
         "Invalid SSH_FXP_STATUS message");
+    sftp_set_error(msg->sftp, SSH_FX_FAILURE);
     return NULL;
   }
   rc = ssh_buffer_unpack(msg->payload, "ss",
@@ -859,6 +925,7 @@ static sftp_status_message parse_status_msg(sftp_message msg){
       SAFE_FREE(status);
       ssh_set_error(msg->sftp->session, SSH_FATAL,
         "Invalid SSH_FXP_STATUS message");
+      sftp_set_error(msg->sftp, SSH_FX_FAILURE);
       return NULL;
   }
   if (status->errormsg == NULL)
@@ -867,6 +934,7 @@ static sftp_status_message parse_status_msg(sftp_message msg){
     status->langmsg = strdup("");
   if (status->errormsg == NULL || status->langmsg == NULL) {
     ssh_set_error_oom(msg->sftp->session);
+    sftp_set_error(msg->sftp, SSH_FX_FAILURE);
     status_msg_free(status);
     return NULL;
   }
@@ -896,6 +964,7 @@ static sftp_file parse_handle_msg(sftp_message msg){
   file = calloc(1, sizeof(struct sftp_file_struct));
   if (file == NULL) {
     ssh_set_error_oom(msg->sftp->session);
+    sftp_set_error(msg->sftp, SSH_FX_FAILURE);
     return NULL;
   }
 
@@ -904,6 +973,7 @@ static sftp_file parse_handle_msg(sftp_message msg){
     ssh_set_error(msg->sftp->session, SSH_FATAL,
         "Invalid SSH_FXP_HANDLE message");
     SAFE_FREE(file);
+    sftp_set_error(msg->sftp, SSH_FX_FAILURE);
     return NULL;
   }
 
@@ -925,9 +995,14 @@ sftp_dir sftp_opendir(sftp_session sftp, const char *path)
     uint32_t id;
     int rc;
 
+    if (sftp == NULL) {
+        return NULL;
+    }
+
     payload = ssh_buffer_new();
     if (payload == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -940,6 +1015,7 @@ sftp_dir sftp_opendir(sftp_session sftp, const char *path)
     if (rc != SSH_OK) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(payload);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -1019,6 +1095,7 @@ static sftp_attributes sftp_parse_attr_4(sftp_session sftp, ssh_buffer buf,
   attr = calloc(1, sizeof(struct sftp_attributes_struct));
   if (attr == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return NULL;
   }
 
@@ -1236,6 +1313,7 @@ static sftp_attributes sftp_parse_attr_3(sftp_session sftp, ssh_buffer buf,
     attr = calloc(1, sizeof(struct sftp_attributes_struct));
     if (attr == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -1246,7 +1324,7 @@ static sftp_attributes sftp_parse_attr_3(sftp_session sftp, ssh_buffer buf,
         if (rc != SSH_OK){
             goto error;
         }
-        SSH_LOG(SSH_LOG_RARE, "Name: %s", attr->name);
+        SSH_LOG(SSH_LOG_PROTOCOL, "Name: %s", attr->name);
 
         /* Set owner and group if we talk to openssh and have the longname */
         if (ssh_get_openssh_version(sftp->session)) {
@@ -1266,17 +1344,17 @@ static sftp_attributes sftp_parse_attr_3(sftp_session sftp, ssh_buffer buf,
     if (rc != SSH_OK){
         goto error;
     }
-    SSH_LOG(SSH_LOG_RARE,
-            "Flags: %.8lx\n", (long unsigned int) attr->flags);
+    SSH_LOG(SSH_LOG_PROTOCOL,
+            "Flags: %.8"PRIx32"\n", (uint32_t) attr->flags);
 
     if (attr->flags & SSH_FILEXFER_ATTR_SIZE) {
         rc = ssh_buffer_unpack(buf, "q", &attr->size);
         if(rc != SSH_OK) {
             goto error;
         }
-        SSH_LOG(SSH_LOG_RARE,
-                "Size: %llu\n",
-                (long long unsigned int) attr->size);
+        SSH_LOG(SSH_LOG_PROTOCOL,
+                "Size: %"PRIu64"\n",
+                (uint64_t) attr->size);
     }
 
     if (attr->flags & SSH_FILEXFER_ATTR_UIDGID) {
@@ -1365,6 +1443,7 @@ static sftp_attributes sftp_parse_attr_3(sftp_session sftp, ssh_buffer buf,
     SAFE_FREE(attr->group);
     SAFE_FREE(attr);
     ssh_set_error(sftp->session, SSH_FATAL, "Invalid ATTR structure");
+    sftp_set_error(sftp, SSH_FX_FAILURE);
 
     return NULL;
 }
@@ -1453,6 +1532,7 @@ sftp_attributes sftp_readdir(sftp_session sftp, sftp_dir dir)
         payload = ssh_buffer_new();
         if (payload == NULL) {
             ssh_set_error_oom(sftp->session);
+            sftp_set_error(sftp, SSH_FX_FAILURE);
             return NULL;
         }
 
@@ -1464,6 +1544,7 @@ sftp_attributes sftp_readdir(sftp_session sftp, sftp_dir dir)
                              dir->handle);
         if (rc != 0) {
             ssh_set_error_oom(sftp->session);
+            sftp_set_error(sftp, SSH_FX_FAILURE);
             ssh_buffer_free(payload);
             return NULL;
         }
@@ -1518,6 +1599,7 @@ sftp_attributes sftp_readdir(sftp_session sftp, sftp_dir dir)
                 ssh_set_error(sftp->session, SSH_FATAL,
                         "Unsupported message back %d", msg->packet_type);
                 sftp_message_free(msg);
+                sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
 
                 return NULL;
         }
@@ -1531,7 +1613,7 @@ sftp_attributes sftp_readdir(sftp_session sftp, sftp_dir dir)
         return NULL;
     }
 
-    SSH_LOG(SSH_LOG_RARE, "Count is %d", dir->count);
+    SSH_LOG(SSH_LOG_PROTOCOL, "Count is %d", dir->count);
 
     attr = sftp_parse_attr(sftp, dir->buffer, 1);
     if (attr == NULL) {
@@ -1583,6 +1665,7 @@ static int sftp_handle_close(sftp_session sftp, ssh_string handle)
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -1595,6 +1678,7 @@ static int sftp_handle_close(sftp_session sftp, ssh_string handle)
     if (rc != SSH_OK) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -1636,6 +1720,7 @@ static int sftp_handle_close(sftp_session sftp, ssh_string handle)
             ssh_set_error(sftp->session, SSH_FATAL,
                     "Received message %d during sftp_handle_close!", msg->packet_type);
             sftp_message_free(msg);
+            sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return -1;
@@ -1725,6 +1810,7 @@ sftp_file sftp_open(sftp_session sftp,
     if (rc != SSH_OK) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -1732,6 +1818,7 @@ sftp_file sftp_open(sftp_session sftp,
     if (rc < 0) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -1776,6 +1863,7 @@ sftp_file sftp_open(sftp_session sftp,
                             SSH_FATAL,
                             "Cannot open in append mode. Unknown file size.");
                     sftp_close(handle);
+                    sftp_set_error(sftp, SSH_FX_FAILURE);
                     return NULL;
                 }
 
@@ -1786,6 +1874,7 @@ sftp_file sftp_open(sftp_session sftp,
             ssh_set_error(sftp->session, SSH_FATAL,
                     "Received message %d during open!", msg->packet_type);
             sftp_message_free(msg);
+            sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return NULL;
@@ -1830,6 +1919,7 @@ ssize_t sftp_read(sftp_file handle, void *buf, size_t count) {
   if (rc != SSH_OK){
     ssh_set_error_oom(sftp->session);
     ssh_buffer_free(buffer);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
   if (sftp_packet_write(handle->sftp, SSH_FXP_READ, buffer) < 0) {
@@ -1898,6 +1988,7 @@ ssize_t sftp_read(sftp_file handle, void *buf, size_t count) {
       ssh_set_error(sftp->session, SSH_FATAL,
           "Received message %d during read!", msg->packet_type);
       sftp_message_free(msg);
+      sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
       return -1;
   }
 
@@ -1914,6 +2005,7 @@ int sftp_async_read_begin(sftp_file file, uint32_t len){
   buffer = ssh_buffer_new();
   if (buffer == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -1928,6 +2020,7 @@ int sftp_async_read_begin(sftp_file file, uint32_t len){
   if (rc != SSH_OK) {
     ssh_set_error_oom(sftp->session);
     ssh_buffer_free(buffer);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
   if (sftp_packet_write(sftp, SSH_FXP_READ, buffer) < 0) {
@@ -2018,6 +2111,7 @@ int sftp_async_read(sftp_file file, void *data, uint32_t size, uint32_t id){
     default:
       ssh_set_error(sftp->session,SSH_FATAL,"Received message %d during read!",msg->packet_type);
       sftp_message_free(msg);
+      sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
       return SSH_ERROR;
   }
 
@@ -2037,6 +2131,7 @@ ssize_t sftp_write(sftp_file file, const void *buf, size_t count) {
   buffer = ssh_buffer_new();
   if (buffer == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -2052,6 +2147,7 @@ ssize_t sftp_write(sftp_file file, const void *buf, size_t count) {
   if (rc != SSH_OK){
     ssh_set_error_oom(sftp->session);
     ssh_buffer_free(buffer);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
   packetlen=ssh_buffer_get_len(buffer);
@@ -2097,6 +2193,7 @@ ssize_t sftp_write(sftp_file file, const void *buf, size_t count) {
       ssh_set_error(sftp->session, SSH_FATAL,
           "Received message %d during write!", msg->packet_type);
       sftp_message_free(msg);
+      sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
       return -1;
   }
 
@@ -2152,6 +2249,7 @@ int sftp_unlink(sftp_session sftp, const char *file) {
   buffer = ssh_buffer_new();
   if (buffer == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -2164,6 +2262,7 @@ int sftp_unlink(sftp_session sftp, const char *file) {
   if (rc != SSH_OK) {
     ssh_set_error_oom(sftp->session);
     ssh_buffer_free(buffer);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -2208,6 +2307,7 @@ int sftp_unlink(sftp_session sftp, const char *file) {
     ssh_set_error(sftp->session,SSH_FATAL,
         "Received message %d when attempting to remove file", msg->packet_type);
     sftp_message_free(msg);
+    sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
   }
 
   return -1;
@@ -2224,6 +2324,7 @@ int sftp_rmdir(sftp_session sftp, const char *directory) {
   buffer = ssh_buffer_new();
   if (buffer == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -2236,6 +2337,7 @@ int sftp_rmdir(sftp_session sftp, const char *directory) {
   if (rc != SSH_OK) {
     ssh_set_error_oom(sftp->session);
     ssh_buffer_free(buffer);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
   if (sftp_packet_write(sftp, SSH_FXP_RMDIR, buffer) < 0) {
@@ -2263,7 +2365,6 @@ int sftp_rmdir(sftp_session sftp, const char *directory) {
       case SSH_FX_OK:
         status_msg_free(status);
         return 0;
-        break;
       default:
         break;
     }
@@ -2276,6 +2377,7 @@ int sftp_rmdir(sftp_session sftp, const char *directory) {
         "Received message %d when attempting to remove directory",
         msg->packet_type);
     sftp_message_free(msg);
+    sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
   }
 
   return -1;
@@ -2295,6 +2397,7 @@ int sftp_mkdir(sftp_session sftp, const char *directory, mode_t mode)
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -2311,6 +2414,7 @@ int sftp_mkdir(sftp_session sftp, const char *directory, mode_t mode)
     if (rc != SSH_OK) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -2318,6 +2422,7 @@ int sftp_mkdir(sftp_session sftp, const char *directory, mode_t mode)
     if (rc < 0) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -2358,7 +2463,6 @@ int sftp_mkdir(sftp_session sftp, const char *directory, mode_t mode)
             case SSH_FX_OK:
                 status_msg_free(status);
                 return 0;
-                break;
             default:
                 break;
         }
@@ -2375,6 +2479,7 @@ int sftp_mkdir(sftp_session sftp, const char *directory, mode_t mode)
                 "Received message %d when attempting to make directory",
                 msg->packet_type);
         sftp_message_free(msg);
+        sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return -1;
@@ -2391,6 +2496,7 @@ int sftp_rename(sftp_session sftp, const char *original, const char *newname) {
   buffer = ssh_buffer_new();
   if (buffer == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -2404,6 +2510,7 @@ int sftp_rename(sftp_session sftp, const char *original, const char *newname) {
   if (rc != SSH_OK) {
     ssh_set_error_oom(sftp->session);
     ssh_buffer_free(buffer);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -2454,6 +2561,7 @@ int sftp_rename(sftp_session sftp, const char *original, const char *newname) {
         "Received message %d when attempting to rename",
         msg->packet_type);
     sftp_message_free(msg);
+    sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
   }
 
   return -1;
@@ -2472,6 +2580,7 @@ int sftp_setstat(sftp_session sftp, const char *file, sftp_attributes attr)
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -2484,6 +2593,7 @@ int sftp_setstat(sftp_session sftp, const char *file, sftp_attributes attr)
     if (rc != SSH_OK) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -2491,6 +2601,7 @@ int sftp_setstat(sftp_session sftp, const char *file, sftp_attributes attr)
     if (rc != 0) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -2534,6 +2645,7 @@ int sftp_setstat(sftp_session sftp, const char *file, sftp_attributes attr)
         ssh_set_error(sftp->session, SSH_FATAL,
                 "Received message %d when attempting to set stats", msg->packet_type);
         sftp_message_free(msg);
+        sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return -1;
@@ -2591,12 +2703,14 @@ int sftp_symlink(sftp_session sftp, const char *target, const char *dest) {
     return -1;
   if (target == NULL || dest == NULL) {
     ssh_set_error_invalid(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
   buffer = ssh_buffer_new();
   if (buffer == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return -1;
   }
 
@@ -2619,6 +2733,7 @@ int sftp_symlink(sftp_session sftp, const char *target, const char *dest) {
   if (rc != SSH_OK){
       ssh_set_error_oom(sftp->session);
       ssh_buffer_free(buffer);
+      sftp_set_error(sftp, SSH_FX_FAILURE);
       return -1;
   }
 
@@ -2662,6 +2777,7 @@ int sftp_symlink(sftp_session sftp, const char *target, const char *dest) {
     ssh_set_error(sftp->session, SSH_FATAL,
         "Received message %d when attempting to set stats", msg->packet_type);
     sftp_message_free(msg);
+    sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
   }
 
   return -1;
@@ -2681,15 +2797,18 @@ char *sftp_readlink(sftp_session sftp, const char *path)
 
     if (path == NULL) {
         ssh_set_error_invalid(sftp);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
     if (sftp->version < 3){
         ssh_set_error(sftp,SSH_REQUEST_DENIED,"sftp version %d does not support sftp_readlink",sftp->version);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -2702,6 +2821,7 @@ char *sftp_readlink(sftp_session sftp, const char *path)
     if (rc < 0) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -2731,6 +2851,7 @@ char *sftp_readlink(sftp_session sftp, const char *path)
             ssh_set_error(sftp->session,
                           SSH_ERROR,
                           "Failed to retrieve link");
+            sftp_set_error(sftp, SSH_FX_FAILURE);
             return NULL;
         }
 
@@ -2741,6 +2862,7 @@ char *sftp_readlink(sftp_session sftp, const char *path)
         if (status == NULL) {
             return NULL;
         }
+        sftp_set_error(sftp, status->status);
         ssh_set_error(sftp->session, SSH_REQUEST_DENIED,
                 "SFTP server: %s", status->errormsg);
         status_msg_free(status);
@@ -2748,6 +2870,7 @@ char *sftp_readlink(sftp_session sftp, const char *path)
         ssh_set_error(sftp->session, SSH_FATAL,
                 "Received message %d when attempting to set stats", msg->packet_type);
         sftp_message_free(msg);
+        sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return NULL;
@@ -2760,6 +2883,7 @@ static sftp_statvfs_t sftp_parse_statvfs(sftp_session sftp, ssh_buffer buf) {
   statvfs = calloc(1, sizeof(struct sftp_statvfs_struct));
   if (statvfs == NULL) {
     ssh_set_error_oom(sftp->session);
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return NULL;
   }
 
@@ -2779,6 +2903,7 @@ static sftp_statvfs_t sftp_parse_statvfs(sftp_session sftp, ssh_buffer buf) {
   if (rc != SSH_OK) {
     SAFE_FREE(statvfs);
     ssh_set_error(sftp->session, SSH_FATAL, "Invalid statvfs structure");
+    sftp_set_error(sftp, SSH_FX_FAILURE);
     return NULL;
   }
 
@@ -2797,16 +2922,19 @@ sftp_statvfs_t sftp_statvfs(sftp_session sftp, const char *path)
         return NULL;
     if (path == NULL) {
         ssh_set_error_invalid(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
     if (sftp->version < 3){
         ssh_set_error(sftp,SSH_REQUEST_DENIED,"sftp version %d does not support sftp_statvfs",sftp->version);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -2820,6 +2948,7 @@ sftp_statvfs_t sftp_statvfs(sftp_session sftp, const char *path)
     if (rc != SSH_OK) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -2850,6 +2979,7 @@ sftp_statvfs_t sftp_statvfs(sftp_session sftp, const char *path)
         if (status == NULL) {
             return NULL;
         }
+        sftp_set_error(sftp, status->status);
         ssh_set_error(sftp->session, SSH_REQUEST_DENIED,
                 "SFTP server: %s", status->errormsg);
         status_msg_free(status);
@@ -2857,6 +2987,7 @@ sftp_statvfs_t sftp_statvfs(sftp_session sftp, const char *path)
         ssh_set_error(sftp->session, SSH_FATAL,
                 "Received message %d when attempting to get statvfs", msg->packet_type);
         sftp_message_free(msg);
+        sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return NULL;
@@ -2878,6 +3009,7 @@ int sftp_fsync(sftp_file file)
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return -1;
     }
 
@@ -2890,6 +3022,7 @@ int sftp_fsync(sftp_file file)
                          file->handle);
     if (rc < 0) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         goto done;
     }
 
@@ -2949,6 +3082,7 @@ int sftp_fsync(sftp_file file)
                       "Received message %d when attempting to set stats",
                       msg->packet_type);
         sftp_message_free(msg);
+        sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     rc = -1;
@@ -2975,6 +3109,7 @@ sftp_statvfs_t sftp_fstatvfs(sftp_file file)
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -2988,6 +3123,7 @@ sftp_statvfs_t sftp_fstatvfs(sftp_file file)
     if (rc < 0) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -3018,6 +3154,7 @@ sftp_statvfs_t sftp_fstatvfs(sftp_file file)
         if (status == NULL) {
             return NULL;
         }
+        sftp_set_error(sftp, status->status);
         ssh_set_error(sftp->session, SSH_REQUEST_DENIED,
                 "SFTP server: %s", status->errormsg);
         status_msg_free(status);
@@ -3025,6 +3162,7 @@ sftp_statvfs_t sftp_fstatvfs(sftp_file file)
         ssh_set_error(sftp->session, SSH_FATAL,
                 "Received message %d when attempting to set stats", msg->packet_type);
         sftp_message_free(msg);
+        sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return NULL;
@@ -3051,12 +3189,14 @@ char *sftp_canonicalize_path(sftp_session sftp, const char *path)
         return NULL;
     if (path == NULL) {
         ssh_set_error_invalid(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -3069,6 +3209,7 @@ char *sftp_canonicalize_path(sftp_session sftp, const char *path)
     if (rc < 0) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -3098,6 +3239,7 @@ char *sftp_canonicalize_path(sftp_session sftp, const char *path)
             ssh_set_error(sftp->session,
                           SSH_ERROR,
                           "Failed to parse canonicalized path");
+            sftp_set_error(sftp, SSH_FX_FAILURE);
             return NULL;
         }
 
@@ -3108,6 +3250,7 @@ char *sftp_canonicalize_path(sftp_session sftp, const char *path)
         if (status == NULL) {
             return NULL;
         }
+        sftp_set_error(sftp, status->status);
         ssh_set_error(sftp->session, SSH_REQUEST_DENIED,
                 "SFTP server: %s", status->errormsg);
         status_msg_free(status);
@@ -3115,6 +3258,7 @@ char *sftp_canonicalize_path(sftp_session sftp, const char *path)
         ssh_set_error(sftp->session, SSH_FATAL,
                 "Received message %d when attempting to set stats", msg->packet_type);
         sftp_message_free(msg);
+        sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
     }
 
     return NULL;
@@ -3130,14 +3274,20 @@ static sftp_attributes sftp_xstat(sftp_session sftp,
     uint32_t id;
     int rc;
 
+    if (sftp == NULL) {
+        return NULL;
+    }
+
     if (path == NULL) {
         ssh_set_error_invalid(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(sftp->session);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -3150,6 +3300,7 @@ static sftp_attributes sftp_xstat(sftp_session sftp,
     if (rc != SSH_OK) {
         ssh_set_error_oom(sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -3186,6 +3337,7 @@ static sftp_attributes sftp_xstat(sftp_session sftp,
     ssh_set_error(sftp->session, SSH_FATAL,
             "Received mesg %d during stat()", msg->packet_type);
     sftp_message_free(msg);
+    sftp_set_error(sftp, SSH_FX_BAD_MESSAGE);
 
     return NULL;
 }
@@ -3206,9 +3358,14 @@ sftp_attributes sftp_fstat(sftp_file file)
     uint32_t id;
     int rc;
 
+    if (file == NULL) {
+        return NULL;
+    }
+
     buffer = ssh_buffer_new();
     if (buffer == NULL) {
         ssh_set_error_oom(file->sftp->session);
+        sftp_set_error(file->sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -3221,6 +3378,7 @@ sftp_attributes sftp_fstat(sftp_file file)
     if (rc != SSH_OK) {
         ssh_set_error_oom(file->sftp->session);
         ssh_buffer_free(buffer);
+        sftp_set_error(file->sftp, SSH_FX_FAILURE);
         return NULL;
     }
 
@@ -3248,6 +3406,7 @@ sftp_attributes sftp_fstat(sftp_file file)
         if (status == NULL) {
             return NULL;
         }
+        sftp_set_error(file->sftp, status->status);
         ssh_set_error(file->sftp->session, SSH_REQUEST_DENIED,
                 "SFTP server: %s", status->errormsg);
         status_msg_free(status);
@@ -3257,6 +3416,7 @@ sftp_attributes sftp_fstat(sftp_file file)
     ssh_set_error(file->sftp->session, SSH_FATAL,
             "Received msg %d during fstat()", msg->packet_type);
     sftp_message_free(msg);
+    sftp_set_error(file->sftp, SSH_FX_BAD_MESSAGE);
 
     return NULL;
 }
