@@ -24,9 +24,11 @@
 #define LIBSSH_STATIC
 
 #include "torture.h"
+#include "torture_key.h"
 
 #include <sys/types.h>
 #include <pwd.h>
+#include <errno.h>
 
 #include "session.c"
 #include "known_hosts.c"
@@ -44,7 +46,7 @@
 
 static int sshd_setup(void **state)
 {
-    torture_setup_sshd_server(state);
+    torture_setup_sshd_server(state, false);
 
     return 0;
 }
@@ -60,10 +62,13 @@ static int session_setup(void **state)
     struct torture_state *s = *state;
     int verbosity = torture_libssh_verbosity();
     struct passwd *pwd;
+    int rc;
 
     pwd = getpwnam("bob");
     assert_non_null(pwd);
-    setuid(pwd->pw_uid);
+
+    rc = setuid(pwd->pw_uid);
+    assert_return_code(rc, errno);
 
     s->ssh.session = ssh_new();
     assert_non_null(s->ssh.session);
@@ -112,22 +117,22 @@ static void torture_knownhosts_port(void **state) {
              TORTURE_KNOWN_HOSTS_FILE);
 
     rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_connect(session);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     session->opts.port = 1234;
     rc = ssh_write_knownhost(session);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     file = fopen(known_hosts_file, "r");
-    assert_true(file != NULL);
+    assert_non_null(file);
     p = fgets(buffer, sizeof(buffer), file);
-    assert_false(p == NULL);
+    assert_non_null(p);
     fclose(file);
     buffer[sizeof(buffer) - 1] = '\0';
-    assert_true(strstr(buffer,"[127.0.0.10]:1234 ") != NULL);
+    assert_non_null(strstr(buffer,"[127.0.0.10]:1234 "));
 
     ssh_disconnect(session);
     ssh_free(session);
@@ -139,13 +144,79 @@ static void torture_knownhosts_port(void **state) {
     ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
 
     rc = ssh_connect(session);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     session->opts.port = 1234;
     rc = ssh_is_server_known(session);
     assert_int_equal(rc, SSH_SERVER_KNOWN_OK);
 }
 
+static void torture_knownhosts_wildcard(void **state)
+{
+    struct torture_state *s = *state;
+    ssh_session session = s->ssh.session;
+    char known_hosts_file[1024];
+    const char *key = NULL;
+    FILE *file;
+    int rc;
+
+    snprintf(known_hosts_file,
+             sizeof(known_hosts_file),
+             "%s/%s",
+             s->socket_dir,
+             TORTURE_KNOWN_HOSTS_FILE);
+
+    file = fopen(known_hosts_file, "w");
+    assert_non_null(file);
+    key = torture_get_testkey_pub(SSH_KEYTYPE_RSA);
+    fprintf(file, "[127.0.0.10]:* %s\n", key);
+    fclose(file);
+
+    rc = ssh_options_set(session, SSH_OPTIONS_HOST, TORTURE_SSH_SERVER);
+    assert_ssh_return_code(session, rc);
+    rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
+    assert_ssh_return_code(session, rc);
+
+    rc = ssh_connect(session);
+    assert_ssh_return_code(session, rc);
+
+    rc = ssh_is_server_known(session);
+    assert_int_equal(rc, SSH_SERVER_KNOWN_OK);
+}
+
+static void torture_knownhosts_standard_port(void **state)
+{
+    struct torture_state *s = *state;
+    ssh_session session = s->ssh.session;
+    char known_hosts_file[1024];
+    const char *key = NULL;
+    FILE *file;
+    int rc;
+
+    snprintf(known_hosts_file,
+             sizeof(known_hosts_file),
+             "%s/%s",
+             s->socket_dir,
+             TORTURE_KNOWN_HOSTS_FILE);
+
+    file = fopen(known_hosts_file, "w");
+    assert_non_null(file);
+    key = torture_get_testkey_pub(SSH_KEYTYPE_RSA);
+    fprintf(file, "[127.0.0.10]:22 %s\n", key);
+    fclose(file);
+
+    rc = ssh_options_set(session, SSH_OPTIONS_HOST, TORTURE_SSH_SERVER);
+    assert_ssh_return_code(session, rc);
+    rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
+    assert_ssh_return_code(session, rc);
+
+    rc = ssh_connect(session);
+    assert_ssh_return_code(session, rc);
+
+    rc = ssh_is_server_known(session);
+    assert_int_equal(rc, SSH_SERVER_KNOWN_OK);
+}
+
 static void torture_knownhosts_fail(void **state) {
     struct torture_state *s = *state;
     ssh_session session = s->ssh.session;
@@ -160,10 +231,10 @@ static void torture_knownhosts_fail(void **state) {
              TORTURE_KNOWN_HOSTS_FILE);
 
     rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
-    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ssh-rsa");
-    assert_int_equal(rc, SSH_OK);
+    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "rsa-sha2-256");
+    assert_ssh_return_code(session, rc);
 
     file = fopen(known_hosts_file, "w");
     assert_non_null(file);
@@ -172,7 +243,7 @@ static void torture_knownhosts_fail(void **state) {
     fclose(file);
 
     rc = ssh_connect(session);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_is_server_known(session);
     assert_int_equal(rc, SSH_SERVER_KNOWN_CHANGED);
@@ -192,10 +263,10 @@ static void torture_knownhosts_other(void **state) {
              TORTURE_KNOWN_HOSTS_FILE);
 
     rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
-    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ssh-ed25519");
-    assert_int_equal(rc, SSH_OK);
+    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ecdsa-sha2-nistp521");
+    assert_ssh_return_code(session, rc);
 
     file = fopen(known_hosts_file, "w");
     assert_non_null(file);
@@ -203,7 +274,7 @@ static void torture_knownhosts_other(void **state) {
     fclose(file);
 
     rc = ssh_connect(session);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_is_server_known(session);
     assert_int_equal(rc, SSH_SERVER_FOUND_OTHER);
@@ -222,22 +293,22 @@ static void torture_knownhosts_other_auto(void **state) {
              TORTURE_KNOWN_HOSTS_FILE);
 
     rc = ssh_options_set(session, SSH_OPTIONS_HOST, TORTURE_SSH_SERVER);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
-    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ssh-ed25519");
-    assert_int_equal(rc, SSH_OK);
+    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ecdsa-sha2-nistp521");
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_connect(session);
-    assert_true(rc==SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_is_server_known(session);
     assert_int_equal(rc, SSH_SERVER_NOT_KNOWN);
 
     rc = ssh_write_knownhost(session);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     ssh_disconnect(session);
     ssh_free(session);
@@ -249,13 +320,13 @@ static void torture_knownhosts_other_auto(void **state) {
     s->ssh.session = session;
 
     rc = ssh_options_set(session, SSH_OPTIONS_HOST, TORTURE_SSH_SERVER);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_connect(session);
-    assert_true(rc==SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     /* ssh-rsa is the default but libssh should try ssh-ed25519 instead */
     rc = ssh_is_server_known(session);
@@ -278,28 +349,28 @@ static void torture_knownhosts_conflict(void **state) {
              TORTURE_KNOWN_HOSTS_FILE);
 
     rc = ssh_options_set(session, SSH_OPTIONS_HOST, TORTURE_SSH_SERVER);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
-    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ssh-rsa");
-    assert_int_equal(rc, SSH_OK);
+    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "rsa-sha2-256");
+    assert_ssh_return_code(session, rc);
 
     file = fopen(known_hosts_file, "w");
-    assert_true(file != NULL);
+    assert_non_null(file);
     fprintf(file, "127.0.0.10 ssh-rsa %s\n", BADRSA);
     fprintf(file, "127.0.0.10 ssh-ed25519 %s\n", BADED25519);
     fclose(file);
 
     rc = ssh_connect(session);
-    assert_true(rc==SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_is_server_known(session);
     assert_int_equal(rc, SSH_SERVER_KNOWN_CHANGED);
 
     rc = ssh_write_knownhost(session);
-    assert_true(rc==SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     ssh_disconnect(session);
     ssh_free(session);
@@ -312,11 +383,11 @@ static void torture_knownhosts_conflict(void **state) {
 
     ssh_options_set(session, SSH_OPTIONS_HOST, TORTURE_SSH_SERVER);
     ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
-    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ssh-rsa");
-    assert_int_equal(rc, SSH_OK);
+    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "rsa-sha2-256");
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_connect(session);
-    assert_int_equal(rc, SSH_OK);
+    assert_ssh_return_code(session, rc);
 
     rc = ssh_is_server_known(session);
     assert_int_equal(rc, SSH_SERVER_KNOWN_OK);
@@ -324,9 +395,50 @@ static void torture_knownhosts_conflict(void **state) {
     /* session will be freed by session_teardown() */
 }
 
+static void torture_knownhosts_no_hostkeychecking(void **state)
+{
+
+    struct torture_state *s = *state;
+    ssh_session session = s->ssh.session;
+    char known_hosts_file[1024] = {0};
+    enum ssh_known_hosts_e found;
+    int strict_host_key_checking = 0;
+    int rc;
+
+    snprintf(known_hosts_file,
+             sizeof(known_hosts_file),
+             "%s/%s",
+             s->socket_dir,
+             TORTURE_KNOWN_HOSTS_FILE);
+
+    rc = ssh_options_set(session, SSH_OPTIONS_KNOWNHOSTS, known_hosts_file);
+    assert_ssh_return_code(session, rc);
+
+    rc = ssh_options_set(session, SSH_OPTIONS_HOSTKEYS, "ecdsa-sha2-nistp521");
+    assert_ssh_return_code(session, rc);
+
+    rc = ssh_connect(session);
+    assert_ssh_return_code(session, rc);
+
+    found = ssh_session_is_known_server(session);
+    assert_int_equal(found, SSH_KNOWN_HOSTS_UNKNOWN);
+
+    rc = ssh_options_set(session, SSH_OPTIONS_STRICTHOSTKEYCHECK, &strict_host_key_checking);
+    assert_ssh_return_code(session, rc);
+
+    found = ssh_session_is_known_server(session);
+    assert_int_equal(found, SSH_KNOWN_HOSTS_OK);
+}
+
 int torture_run_tests(void) {
     int rc;
     struct CMUnitTest tests[] = {
+        cmocka_unit_test_setup_teardown(torture_knownhosts_wildcard,
+                                        session_setup,
+                                        session_teardown),
+        cmocka_unit_test_setup_teardown(torture_knownhosts_standard_port,
+                                        session_setup,
+                                        session_teardown),
         cmocka_unit_test_setup_teardown(torture_knownhosts_port,
                                         session_setup,
                                         session_teardown),
@@ -342,6 +454,9 @@ int torture_run_tests(void) {
         cmocka_unit_test_setup_teardown(torture_knownhosts_conflict,
                                         session_setup,
                                         session_teardown),
+        cmocka_unit_test_setup_teardown(torture_knownhosts_no_hostkeychecking,
+                                        session_setup,
+                                        session_teardown),
     };
 
     ssh_init();
