@@ -12,6 +12,7 @@
 #define LIBSSH_ED25519_TESTKEY "libssh_testkey.id_ed25519"
 #define LIBSSH_ED25519_TESTKEY_PASSPHRASE "libssh_testkey_passphrase.id_ed25519"
 
+const char template[] = "temp_dir_XXXXXX";
 const unsigned char HASH[] = "12345678901234567890";
 const uint8_t ref_signature[ED25519_SIG_LEN]=
     "\xbb\x8d\x55\x9f\x06\x14\x39\x24\xb4\xe1\x5a\x57\x3d\x9d\xbe\x22"
@@ -19,50 +20,120 @@ const uint8_t ref_signature[ED25519_SIG_LEN]=
     "\x10\xa3\x18\x93\xdf\xa4\x96\x81\x11\x8e\x1e\x26\x14\x8a\x08\x1b"
     "\x01\x6a\x60\x59\x9c\x4a\x55\xa3\x16\x56\xf6\xc4\x50\x42\x7f\x03";
 
+struct pki_st {
+    char *cwd;
+    char *temp_dir;
+};
+
 static int setup_ed25519_key(void **state)
 {
-    (void) state; /* unused */
+    const char *keystring = NULL;
+    struct pki_st *test_state = NULL;
+    char *cwd = NULL;
+    char *tmp_dir = NULL;
+    int rc = 0;
 
-    unlink(LIBSSH_ED25519_TESTKEY);
-    unlink(LIBSSH_ED25519_TESTKEY_PASSPHRASE);
-    unlink(LIBSSH_ED25519_TESTKEY ".pub");
+    test_state = (struct pki_st *)malloc(sizeof(struct pki_st));
+    assert_non_null(test_state);
 
-    torture_write_file(LIBSSH_ED25519_TESTKEY,
-                       torture_get_testkey(SSH_KEYTYPE_ED25519, 0,0));
-    torture_write_file(LIBSSH_ED25519_TESTKEY_PASSPHRASE,
-                       torture_get_testkey(SSH_KEYTYPE_ED25519, 0,0));
+    cwd = torture_get_current_working_dir();
+    assert_non_null(cwd);
+
+    tmp_dir = torture_make_temp_dir(template);
+    assert_non_null(tmp_dir);
+
+    test_state->cwd = cwd;
+    test_state->temp_dir = tmp_dir;
+
+    *state = test_state;
+
+    rc = torture_change_dir(tmp_dir);
+    assert_int_equal(rc, 0);
+
+    printf("Changed directory to: %s\n", tmp_dir);
+
+    keystring = torture_get_openssh_testkey(SSH_KEYTYPE_ED25519, 0);
+    torture_write_file(LIBSSH_ED25519_TESTKEY, keystring);
+    keystring = torture_get_openssh_testkey(SSH_KEYTYPE_ED25519, 1);
+    torture_write_file(LIBSSH_ED25519_TESTKEY_PASSPHRASE, keystring);
 
     torture_write_file(LIBSSH_ED25519_TESTKEY ".pub",
-                       torture_get_testkey_pub(SSH_KEYTYPE_ED25519,0));
+                       torture_get_testkey_pub(SSH_KEYTYPE_ED25519));
+    torture_write_file(LIBSSH_ED25519_TESTKEY "-cert.pub",
+                       torture_get_testkey_pub(SSH_KEYTYPE_ED25519_CERT01));
 
     return 0;
 }
 
 static int teardown(void **state) {
-    (void) state; /* unused */
+    struct pki_st *test_state = NULL;
+    int rc = 0;
 
-    unlink(LIBSSH_ED25519_TESTKEY);
-    unlink(LIBSSH_ED25519_TESTKEY_PASSPHRASE);
-    unlink(LIBSSH_ED25519_TESTKEY ".pub");
+    test_state = *((struct pki_st **)state);
+
+    assert_non_null(test_state);
+    assert_non_null(test_state->cwd);
+    assert_non_null(test_state->temp_dir);
+
+    rc = torture_change_dir(test_state->cwd);
+    assert_int_equal(rc, 0);
+
+    rc = torture_rmdirs(test_state->temp_dir);
+    assert_int_equal(rc, 0);
+
+    SAFE_FREE(test_state->temp_dir);
+    SAFE_FREE(test_state->cwd);
+    SAFE_FREE(test_state);
 
     return 0;
 }
 
+static void torture_pki_ed25519_import_pubkey_file(void **state)
+{
+    ssh_key pubkey = NULL;
+    int rc;
+
+    (void)state;
+
+    /* The key doesn't have the hostname as comment after the key */
+    rc = ssh_pki_import_pubkey_file(LIBSSH_ED25519_TESTKEY ".pub", &pubkey);
+    assert_return_code(rc, errno);
+    assert_non_null(pubkey);
+
+    SSH_KEY_FREE(pubkey);
+}
+
+static void torture_pki_ed25519_import_pubkey_from_openssh_privkey(void **state)
+{
+    ssh_key pubkey = NULL;
+    int rc;
+
+    (void)state;
+
+    /* The key doesn't have the hostname as comment after the key */
+    rc = ssh_pki_import_pubkey_file(LIBSSH_ED25519_TESTKEY_PASSPHRASE, &pubkey);
+    assert_return_code(rc, errno);
+    assert_non_null(pubkey);
+
+    SSH_KEY_FREE(pubkey);
+}
+
 static void torture_pki_ed25519_import_privkey_base64(void **state)
 {
     int rc;
-    char *key_str;
-    ssh_key key;
+    char *key_str = NULL;
+    ssh_key key = NULL;
     const char *passphrase = torture_get_testkey_passphrase();
     enum ssh_keytypes_e type;
 
     (void) state; /* unused */
 
     key_str = torture_pki_read_file(LIBSSH_ED25519_TESTKEY);
-    assert_true(key_str != NULL);
+    assert_non_null(key_str);
 
     rc = ssh_pki_import_privkey_base64(key_str, passphrase, NULL, NULL, &key);
     assert_true(rc == 0);
+    assert_non_null(key);
 
     type = ssh_key_type(key);
     assert_true(type == SSH_KEYTYPE_ED25519);
@@ -74,78 +145,161 @@ static void torture_pki_ed25519_import_privkey_base64(void **state)
     assert_true(rc == 1);
 
     free(key_str);
-    ssh_key_free(key);
+    SSH_KEY_FREE(key);
 
 }
 
+static void torture_pki_ed25519_import_export_privkey_base64(void **state)
+{
+    char *b64_key = NULL;
+    ssh_key key = NULL;
+    const char *passphrase = torture_get_testkey_passphrase();
+    enum ssh_keytypes_e type;
+    int rc;
+
+    (void) state; /* unused */
+
+    rc = ssh_pki_import_privkey_base64(torture_get_openssh_testkey(SSH_KEYTYPE_ED25519,
+                                                                   false),
+                                       passphrase,
+                                       NULL,
+                                       NULL,
+                                       &key);
+    assert_return_code(rc, errno);
+    assert_non_null(key);
+
+    type = ssh_key_type(key);
+    assert_true(type == SSH_KEYTYPE_ED25519);
+
+    rc = ssh_key_is_private(key);
+    assert_true(rc == 1);
+
+    rc = ssh_pki_export_privkey_base64(key,
+                                       passphrase,
+                                       NULL,
+                                       NULL,
+                                       &b64_key);
+    assert_return_code(rc, errno);
+    assert_non_null(b64_key);
+    SSH_KEY_FREE(key);
+
+    rc = ssh_pki_import_privkey_base64(b64_key,
+                                       passphrase,
+                                       NULL,
+                                       NULL,
+                                       &key);
+    assert_return_code(rc, errno);
+    assert_non_null(key);
+
+    type = ssh_key_type(key);
+    assert_true(type == SSH_KEYTYPE_ED25519);
+
+    rc = ssh_key_is_private(key);
+    assert_true(rc == 1);
+
+    SSH_STRING_FREE_CHAR(b64_key);
+    SSH_KEY_FREE(key);
+}
+
 static void torture_pki_ed25519_publickey_from_privatekey(void **state)
 {
     int rc;
-    ssh_key key;
-    ssh_key pubkey;
+    ssh_key key = NULL;
+    ssh_key pubkey = NULL;
     const char *passphrase = NULL;
+    const char *keystring = NULL;
 
     (void) state; /* unused */
 
-    rc = ssh_pki_import_privkey_base64(torture_get_testkey(SSH_KEYTYPE_ED25519, 0, 0),
+    keystring = torture_get_openssh_testkey(SSH_KEYTYPE_ED25519, 0);
+    rc = ssh_pki_import_privkey_base64(keystring,
                                        passphrase,
                                        NULL,
                                        NULL,
                                        &key);
     assert_true(rc == 0);
+    assert_non_null(key);
 
     rc = ssh_key_is_private(key);
     assert_true(rc == 1);
 
     rc = ssh_pki_export_privkey_to_pubkey(key, &pubkey);
     assert_true(rc == SSH_OK);
+    assert_non_null(pubkey);
 
-    ssh_key_free(key);
-    ssh_key_free(pubkey);
+    SSH_KEY_FREE(key);
+    SSH_KEY_FREE(pubkey);
+}
+
+static void torture_pki_ed25519_import_cert_file(void **state)
+{
+    int rc;
+    ssh_key cert = NULL;
+    enum ssh_keytypes_e type;
+
+    (void) state; /* unused */
+
+    rc = ssh_pki_import_cert_file(LIBSSH_ED25519_TESTKEY "-cert.pub", &cert);
+    assert_true(rc == 0);
+    assert_non_null(cert);
+
+    type = ssh_key_type(cert);
+    assert_true(type == SSH_KEYTYPE_ED25519_CERT01);
+
+    rc = ssh_key_is_public(cert);
+    assert_true(rc == 1);
+
+    SSH_KEY_FREE(cert);
 }
 
 static void torture_pki_ed25519_publickey_base64(void **state)
 {
     enum ssh_keytypes_e type;
-    char *b64_key, *key_buf, *p;
-    const char *q;
-    ssh_key key;
+    char *b64_key = NULL, *key_buf = NULL, *p = NULL;
+    const char *q = NULL;
+    ssh_key key = NULL;
     int rc;
 
     (void) state; /* unused */
 
-    key_buf = strdup(torture_get_testkey_pub(SSH_KEYTYPE_ED25519, 0));
-    assert_true(key_buf != NULL);
+    key_buf = strdup(torture_get_testkey_pub(SSH_KEYTYPE_ED25519));
+    assert_non_null(key_buf);
 
     q = p = key_buf;
-    while (*p != ' ') p++;
-    *p = '\0';
+    while (p != NULL && *p != '\0' && *p != ' ') p++;
+    if (p != NULL) {
+        *p = '\0';
+    }
 
     type = ssh_key_type_from_name(q);
     assert_true(type == SSH_KEYTYPE_ED25519);
 
     q = ++p;
-    while (*p != ' ') p++;
-    *p = '\0';
+    while (p != NULL && *p != '\0' && *p != ' ') p++;
+    if (p != NULL) {
+        *p = '\0';
+    }
 
     rc = ssh_pki_import_pubkey_base64(q, type, &key);
     assert_true(rc == 0);
+    assert_non_null(key);
 
     rc = ssh_pki_export_pubkey_base64(key, &b64_key);
     assert_true(rc == 0);
+    assert_non_null(b64_key);
 
     assert_string_equal(q, b64_key);
 
     free(b64_key);
     free(key_buf);
-    ssh_key_free(key);
+    SSH_KEY_FREE(key);
 }
 
 static void torture_pki_ed25519_generate_pubkey_from_privkey(void **state)
 {
     char pubkey_generated[4096] = {0};
-    ssh_key privkey;
-    ssh_key pubkey;
+    ssh_key privkey = NULL;
+    ssh_key pubkey = NULL;
     int rc;
     int len;
 
@@ -163,6 +317,7 @@ static void torture_pki_ed25519_generate_pubkey_from_privkey(void **state)
 
     rc = ssh_pki_export_privkey_to_pubkey(privkey, &pubkey);
     assert_true(rc == SSH_OK);
+    assert_non_null(pubkey);
 
     rc = ssh_pki_export_pubkey_file(pubkey, LIBSSH_ED25519_TESTKEY ".pub");
     assert_true(rc == 0);
@@ -172,20 +327,20 @@ static void torture_pki_ed25519_generate_pubkey_from_privkey(void **state)
                                sizeof(pubkey_generated));
     assert_true(rc == 0);
 
-    len = torture_pubkey_len(torture_get_testkey_pub(SSH_KEYTYPE_ED25519, 0));
-    assert_memory_equal(torture_get_testkey_pub(SSH_KEYTYPE_ED25519, 0),
+    len = torture_pubkey_len(torture_get_testkey_pub(SSH_KEYTYPE_ED25519));
+    assert_memory_equal(torture_get_testkey_pub(SSH_KEYTYPE_ED25519),
                         pubkey_generated,
                         len);
 
-    ssh_key_free(privkey);
-    ssh_key_free(pubkey);
+    SSH_KEY_FREE(privkey);
+    SSH_KEY_FREE(pubkey);
 }
 
 static void torture_pki_ed25519_generate_key(void **state)
 {
     int rc;
-    ssh_key key;
-    ssh_signature sign;
+    ssh_key key = NULL;
+    ssh_signature sign = NULL;
     enum ssh_keytypes_e type = SSH_KEYTYPE_UNKNOWN;
     const char *type_char = NULL;
     ssh_session session=ssh_new();
@@ -193,9 +348,9 @@ static void torture_pki_ed25519_generate_key(void **state)
 
     rc = ssh_pki_generate(SSH_KEYTYPE_ED25519, 256, &key);
     assert_true(rc == SSH_OK);
-    assert_true(key != NULL);
-    sign = pki_do_sign(key, HASH, 20);
-    assert_true(sign != NULL);
+    assert_non_null(key);
+    sign = pki_do_sign(key, HASH, 20, SSH_DIGEST_AUTO);
+    assert_non_null(sign);
     rc = pki_signature_verify(session,sign,key,HASH,20);
     assert_true(rc == SSH_OK);
     type = ssh_key_type(key);
@@ -209,16 +364,46 @@ static void torture_pki_ed25519_generate_key(void **state)
     assert_true(rc == SSH_ERROR);
 
     ssh_signature_free(sign);
-    ssh_key_free(key);
-    key=NULL;
+    SSH_KEY_FREE(key);
+
+    ssh_free(session);
+}
+
+static void torture_pki_ed25519_cert_verify(void **state)
+{
+    int rc;
+    ssh_key privkey = NULL, cert = NULL;
+    ssh_signature sign = NULL;
+    ssh_session session=ssh_new();
+    (void) state;
+
+    rc = ssh_pki_import_privkey_file(LIBSSH_ED25519_TESTKEY,
+                                     NULL,
+                                     NULL,
+                                     NULL,
+                                     &privkey);
+    assert_true(rc == 0);
+    assert_non_null(privkey);
+
+    rc = ssh_pki_import_cert_file(LIBSSH_ED25519_TESTKEY "-cert.pub", &cert);
+    assert_true(rc == 0);
+    assert_non_null(cert);
+
+    sign = pki_do_sign(privkey, HASH, 20, SSH_DIGEST_AUTO);
+    assert_non_null(sign);
+    rc = pki_signature_verify(session, sign, cert, HASH, 20);
+    assert_true(rc == SSH_OK);
+    ssh_signature_free(sign);
+    SSH_KEY_FREE(privkey);
+    SSH_KEY_FREE(cert);
 
     ssh_free(session);
 }
 
 static void torture_pki_ed25519_write_privkey(void **state)
 {
-    ssh_key origkey;
-    ssh_key privkey;
+    ssh_key origkey = NULL;
+    ssh_key privkey = NULL;
     int rc;
 
     (void) state; /* unused */
@@ -229,6 +414,7 @@ static void torture_pki_ed25519_write_privkey(void **state)
             NULL,
             &origkey);
     assert_true(rc == 0);
+    assert_non_null(origkey);
 
     unlink(LIBSSH_ED25519_TESTKEY);
 
@@ -245,12 +431,13 @@ static void torture_pki_ed25519_write_privkey(void **state)
             NULL,
             &privkey);
     assert_true(rc == 0);
+    assert_non_null(privkey);
 
     rc = ssh_key_cmp(origkey, privkey, SSH_KEY_CMP_PRIVATE);
     assert_true(rc == 0);
 
     unlink(LIBSSH_ED25519_TESTKEY);
-    ssh_key_free(privkey);
+    SSH_KEY_FREE(privkey);
     /* do the same with passphrase */
     rc = ssh_pki_export_privkey_file(origkey,
             torture_get_testkey_passphrase(),
@@ -273,13 +460,14 @@ static void torture_pki_ed25519_write_privkey(void **state)
             NULL,
             &privkey);
     assert_true(rc == 0);
+    assert_non_null(privkey);
 
     rc = ssh_key_cmp(origkey, privkey, SSH_KEY_CMP_PRIVATE);
     assert_true(rc == 0);
     unlink(LIBSSH_ED25519_TESTKEY);
 
-    ssh_key_free(origkey);
-    ssh_key_free(privkey);
+    SSH_KEY_FREE(origkey);
+    SSH_KEY_FREE(privkey);
 
     /* Test with passphrase */
     rc = ssh_pki_import_privkey_file(LIBSSH_ED25519_TESTKEY_PASSPHRASE,
@@ -288,6 +476,7 @@ static void torture_pki_ed25519_write_privkey(void **state)
                                      NULL,
                                      &origkey);
     assert_true(rc == 0);
+    assert_non_null(origkey);
 
     unlink(LIBSSH_ED25519_TESTKEY_PASSPHRASE);
     rc = ssh_pki_export_privkey_file(origkey,
@@ -311,18 +500,21 @@ static void torture_pki_ed25519_write_privkey(void **state)
                                      NULL,
                                      &privkey);
     assert_true(rc == 0);
+    assert_non_null(privkey);
 
     rc = ssh_key_cmp(origkey, privkey, SSH_KEY_CMP_PRIVATE);
     assert_true(rc == 0);
 
-    ssh_key_free(origkey);
-    ssh_key_free(privkey);
+    SSH_KEY_FREE(origkey);
+    SSH_KEY_FREE(privkey);
 }
 
-static void torture_pki_ed25519_sign(void **state){
-    ssh_key privkey;
-    ssh_signature sig;
-    ssh_string blob;
+static void torture_pki_ed25519_sign(void **state)
+{
+    ssh_key privkey = NULL;
+    ssh_signature sig = NULL;
+    ssh_string blob = NULL;
+    const char *keystring = NULL;
     int rc;
 
     (void)state;
@@ -330,9 +522,8 @@ static void torture_pki_ed25519_sign(void **state){
     sig = ssh_signature_new();
     assert_non_null(sig);
 
-    rc = ssh_pki_import_privkey_base64(torture_get_testkey(SSH_KEYTYPE_ED25519,
-                                                           0,
-                                                           0),
+    keystring = torture_get_openssh_testkey(SSH_KEYTYPE_ED25519, 0);
+    rc = ssh_pki_import_privkey_base64(keystring,
                                        NULL,
                                        NULL,
                                        NULL,
@@ -345,23 +536,23 @@ static void torture_pki_ed25519_sign(void **state){
     assert_true(rc == SSH_OK);
 
     blob = pki_signature_to_blob(sig);
-    assert_true(blob != NULL);
+    assert_non_null(blob);
 
     assert_int_equal(ssh_string_len(blob), sizeof(ref_signature));
     assert_memory_equal(ssh_string_data(blob), ref_signature, sizeof(ref_signature));
     /* ssh_print_hexa("signature", ssh_string_data(blob), ssh_string_len(blob)); */
     ssh_signature_free(sig);
-    ssh_key_free(privkey);
-    ssh_string_free(blob);
+    SSH_KEY_FREE(privkey);
+    SSH_STRING_FREE(blob);
 
 }
 
 static void torture_pki_ed25519_verify(void **state){
-    ssh_key pubkey;
-    ssh_signature sig;
+    ssh_key pubkey = NULL;
+    ssh_signature sig = NULL;
     ssh_string blob = ssh_string_new(ED25519_SIG_LEN);
-    char *pkey_ptr = strdup(strchr(torture_get_testkey_pub(SSH_KEYTYPE_ED25519,0), ' ') + 1);
-    char *ptr;
+    char *pkey_ptr = strdup(strchr(torture_get_testkey_pub(SSH_KEYTYPE_ED25519), ' ') + 1);
+    char *ptr = NULL;
     int rc;
     (void) state;
 
@@ -372,10 +563,11 @@ static void torture_pki_ed25519_verify(void **state){
     }
     rc = ssh_pki_import_pubkey_base64(pkey_ptr, SSH_KEYTYPE_ED25519, &pubkey);
     assert_true(rc == SSH_OK);
+    assert_non_null(pubkey);
 
     ssh_string_fill(blob, ref_signature, ED25519_SIG_LEN);
-    sig = pki_signature_from_blob(pubkey, blob, SSH_KEYTYPE_ED25519);
-    assert_true(sig != NULL);
+    sig = pki_signature_from_blob(pubkey, blob, SSH_KEYTYPE_ED25519, SSH_DIGEST_AUTO);
+    assert_non_null(sig);
 
     rc = pki_ed25519_verify(pubkey, sig, HASH, sizeof(HASH));
     assert_true(rc == SSH_OK);
@@ -383,17 +575,17 @@ static void torture_pki_ed25519_verify(void **state){
     ssh_signature_free(sig);
     /* alter signature and expect false result */
 
-    ssh_key_free(pubkey);
-    ssh_string_free(blob);
+    SSH_KEY_FREE(pubkey);
+    SSH_STRING_FREE(blob);
     free(pkey_ptr);
 }
 
 static void torture_pki_ed25519_verify_bad(void **state){
-    ssh_key pubkey;
-    ssh_signature sig;
+    ssh_key pubkey = NULL;
+    ssh_signature sig = NULL;
     ssh_string blob = ssh_string_new(ED25519_SIG_LEN);
-    char *pkey_ptr = strdup(strchr(torture_get_testkey_pub(SSH_KEYTYPE_ED25519,0), ' ') + 1);
-    char *ptr;
+    char *pkey_ptr = strdup(strchr(torture_get_testkey_pub(SSH_KEYTYPE_ED25519), ' ') + 1);
+    char *ptr = NULL;
     int rc;
     int i;
     (void) state;
@@ -405,22 +597,23 @@ static void torture_pki_ed25519_verify_bad(void **state){
     }
     rc = ssh_pki_import_pubkey_base64(pkey_ptr, SSH_KEYTYPE_ED25519, &pubkey);
     assert_true(rc == SSH_OK);
+    assert_non_null(pubkey);
 
     /* alter signature and expect false result */
 
     for (i=0; i < ED25519_SIG_LEN; ++i){
         ssh_string_fill(blob, ref_signature, ED25519_SIG_LEN);
         ((uint8_t *)ssh_string_data(blob))[i] ^= 0xff;
-        sig = pki_signature_from_blob(pubkey, blob, SSH_KEYTYPE_ED25519);
-        assert_true(sig != NULL);
+        sig = pki_signature_from_blob(pubkey, blob, SSH_KEYTYPE_ED25519, SSH_DIGEST_AUTO);
+        assert_non_null(sig);
 
         rc = pki_ed25519_verify(pubkey, sig, HASH, sizeof(HASH));
         assert_true(rc == SSH_ERROR);
         ssh_signature_free(sig);
 
     }
-    ssh_key_free(pubkey);
-    ssh_string_free(blob);
+    SSH_KEY_FREE(pubkey);
+    SSH_STRING_FREE(blob);
     free(pkey_ptr);
 }
 
@@ -429,31 +622,33 @@ static void torture_pki_ed25519_import_privkey_base64_passphrase(void **state)
     int rc;
     ssh_key key = NULL;
     const char *passphrase = torture_get_testkey_passphrase();
+    const char *testkey = NULL;
 
     (void) state; /* unused */
 
     /* same for ED25519 */
-    rc = ssh_pki_import_privkey_base64(torture_get_testkey(SSH_KEYTYPE_ED25519, 0, 1),
+    testkey = torture_get_openssh_testkey(SSH_KEYTYPE_ED25519, 1);
+    rc = ssh_pki_import_privkey_base64(testkey,
                                        passphrase,
                                        NULL,
                                        NULL,
                                        &key);
     assert_true(rc == 0);
+    assert_non_null(key);
 
     rc = ssh_key_is_private(key);
     assert_true(rc == 1);
 
-    ssh_key_free(key);
-    key = NULL;
+    SSH_KEY_FREE(key);
 
     /* test if it returns -1 if passphrase is wrong */
-    rc = ssh_pki_import_privkey_base64(torture_get_testkey(SSH_KEYTYPE_ED25519, 0, 1),
+    rc = ssh_pki_import_privkey_base64(testkey,
                                        "wrong passphrase !!",
                                        NULL,
                                        NULL,
                                        &key);
     assert_true(rc == -1);
-    ssh_key_free(key);
+    SSH_KEY_FREE(key);
 }
 
 static void torture_pki_ed25519_privkey_dup(void **state)
@@ -461,16 +656,19 @@ static void torture_pki_ed25519_privkey_dup(void **state)
     const char *passphrase = torture_get_testkey_passphrase();
     ssh_key key = NULL;
     ssh_key dup = NULL;
+    const char *testkey = NULL;
     int rc;
 
     (void) state; /* unused */
 
-    rc = ssh_pki_import_privkey_base64(torture_get_testkey(SSH_KEYTYPE_ED25519, 0, 1),
+    testkey = torture_get_openssh_testkey(SSH_KEYTYPE_ED25519, 1);
+    rc = ssh_pki_import_privkey_base64(testkey,
                                        passphrase,
                                        NULL,
                                        NULL,
                                        &key);
     assert_true(rc == 0);
+    assert_non_null(key);
 
     rc = ssh_key_is_private(key);
     assert_true(rc == 1);
@@ -478,15 +676,15 @@ static void torture_pki_ed25519_privkey_dup(void **state)
     dup = ssh_key_dup(key);
     assert_non_null(dup);
 
-    ssh_key_free(key);
-    ssh_key_free(dup);
+    SSH_KEY_FREE(key);
+    SSH_KEY_FREE(dup);
 }
 
 static void torture_pki_ed25519_pubkey_dup(void **state)
 {
     ssh_key pubkey = NULL;
     ssh_key dup = NULL;
-    const char *p = strchr(torture_get_testkey_pub(SSH_KEYTYPE_ED25519, 0), ' ');
+    const char *p = strchr(torture_get_testkey_pub(SSH_KEYTYPE_ED25519), ' ');
     char *pub_str = NULL;
     char *q = NULL;
     int rc;
@@ -504,6 +702,7 @@ static void torture_pki_ed25519_pubkey_dup(void **state)
                                       SSH_KEYTYPE_ED25519,
                                       &pubkey);
     assert_true(rc == 0);
+    assert_non_null(pubkey);
 
     rc = ssh_key_is_public(pubkey);
     assert_true(rc == 1);
@@ -515,19 +714,31 @@ static void torture_pki_ed25519_pubkey_dup(void **state)
     assert_true(rc == 1);
 
     SAFE_FREE(pub_str);
-    ssh_key_free(pubkey);
-    ssh_key_free(dup);
+    SSH_KEY_FREE(pubkey);
+    SSH_KEY_FREE(dup);
 }
 
 int torture_run_tests(void) {
     int rc;
-    const struct CMUnitTest tests[] = {
+    struct CMUnitTest tests[] = {
+        cmocka_unit_test_setup_teardown(torture_pki_ed25519_import_pubkey_file,
+                                        setup_ed25519_key,
+                                        teardown),
+        cmocka_unit_test_setup_teardown(torture_pki_ed25519_import_pubkey_from_openssh_privkey,
+                                        setup_ed25519_key,
+                                        teardown),
         cmocka_unit_test_setup_teardown(torture_pki_ed25519_import_privkey_base64,
                                         setup_ed25519_key,
                                         teardown),
+        cmocka_unit_test_setup_teardown(torture_pki_ed25519_import_export_privkey_base64,
+                                        setup_ed25519_key,
+                                        teardown),
         cmocka_unit_test_setup_teardown(torture_pki_ed25519_publickey_from_privatekey,
                                         setup_ed25519_key,
                                         teardown),
+        cmocka_unit_test_setup_teardown(torture_pki_ed25519_import_cert_file,
+                                        setup_ed25519_key,
+                                        teardown),
         cmocka_unit_test_setup_teardown(torture_pki_ed25519_publickey_base64,
                                         setup_ed25519_key,
                                         teardown),
@@ -535,6 +746,9 @@ int torture_run_tests(void) {
                                         setup_ed25519_key,
                                         teardown),
         cmocka_unit_test(torture_pki_ed25519_generate_key),
+        cmocka_unit_test_setup_teardown(torture_pki_ed25519_cert_verify,
+                                        setup_ed25519_key,
+                                        teardown),
         cmocka_unit_test_setup_teardown(torture_pki_ed25519_write_privkey,
                                         setup_ed25519_key,
                                         teardown),
@@ -547,6 +761,7 @@ int torture_run_tests(void) {
     };
 
     ssh_init();
+    torture_filter_tests(tests);
     rc = cmocka_run_group_tests(tests, NULL, NULL);
     ssh_finalize();
     return rc;
